from keras_cv_attention_models.gcvit.gcvit import GCViT, GCViT_XXTiny, GCViT_XTiny, GCViT_Tiny, GCViT_Small, GCViT_Base

__head_doc__ = """
Keras implementation of [Github NVlabs/GCVit](https://github.com/NVlabs/GCVit).
Paper [PDF 2206.09959 Global Context Vision Transformers](https://arxiv.org/pdf/2206.09959.pdf).
"""

__tail_doc__ = """  window_ratios: window split ratio. Each stack will calculate `window_size = (height // window_ratio, width // window_ratio)` .
  layer_scale: layer scale init value. Default `-1` means not applying, any value `>=0` will add a scale value for each block output.
      [Going deeper with Image Transformers](https://arxiv.org/pdf/2103.17239.pdf).
  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `gelu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be value like `0.2`, indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  dropout: dropout rate if top layers is included.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
  pretrained: one of None or "imagenet".
      Will try to download and load pre-trained model weights if not None.

Returns:
    A `keras.Model` instance.
"""

GCViT.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  num_heads: num heads for each stack.
  embed_dim: basic hidden dims, expand * 2 for each stack.
  mlp_ratio: expand ratio for mlp blocks hidden channel.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model        | Params | FLOPs | Input | Top1 Acc |
  | ------------ | ------ | ----- | ----- | -------- |
  | GCViT_XXTiny | 12.0M  | 2.15G | 224   | 79.8     |
  | GCViT_XTiny  | 20.0M  | 2.96G | 224   | 82.04    |
  | GCViT_Tiny   | 28.2M  | 4.83G | 224   | 83.4     |
  | GCViT_Small  | 51.1M  | 8.63G | 224   | 83.95    |
  | GCViT_Base   | 90.3M  | 14.9G | 224   | 84.47    |
"""

GCViT_XXTiny.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

GCViT_XTiny.__doc__ = GCViT_XXTiny.__doc__
GCViT_Tiny.__doc__ = GCViT_XXTiny.__doc__
GCViT_Small.__doc__ = GCViT_XXTiny.__doc__
GCViT_Base.__doc__ = GCViT_XXTiny.__doc__



from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, image_data_format
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    batchnorm_with_activation,
    conv2d_no_bias,
    depthwise_conv2d_no_bias,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

PRETRAINED_DICT = {
    "vanillanet_10_deploy": {"imagenet": "554b5bbd83141d83fe81d8a86e6d197f"},
    "vanillanet_10": {"imagenet": "ad0e864e059c0681c8a828819f30e8be"},
    "vanillanet_5_deploy": {"imagenet": "39df8736f61ef9536d33be025849a1ac"},
    "vanillanet_5": {"imagenet": "1dcbdec8d9c9a0b7c1a526528bea2e54"},
    "vanillanet_6_deploy": {"imagenet": "8d53cc4b8a9e82347b63fc839766a3bd"},
    "vanillanet_6": {"imagenet": "68d3c1c57711f806ced2cbc410c5d76f"},
    "vanillanet_7_deploy": {"imagenet": "cefc8bf1ec4507a2abe9d7de69245608"},
    "vanillanet_7": {"imagenet": "f2f168fcc59fc69db9a67dee6c23afad"},
    "vanillanet_8_deploy": {"imagenet": "b9deae326d5c0888c9c4a365ea1e99ad"},
    "vanillanet_8": {"imagenet": "20090c6bdef74e86754fcea96ea7ffc9"},
    "vanillanet_9_deploy": {"imagenet": "ef2b73dee1a0ce45865178ae0fcd4656"},
    "vanillanet_9": {"imagenet": "922fb62b1cb81786c14845b020f70082"},
}


BATCH_NORM_EPSILON = 1e-6


def activation_depthwise_conv_bn(inputs, kernel_size=7, deploy=False, activation="relu", name=""):
    nn = activation_by_name(inputs, activation=activation, name=name)
    nn = depthwise_conv2d_no_bias(nn, kernel_size=kernel_size, use_bias=deploy, padding="same", name=name)
    if not deploy:
        nn = batchnorm_with_activation(nn, epsilon=BATCH_NORM_EPSILON, activation=None, name=name)
    return nn


def VanillaNet(
    out_channels=[1024, 2048, 4096],
    strides=[2, 2, 2],
    stem_width=512,
    leaky_relu_alpha=1.0,
    input_shape=(224, 224, 3),
    deploy=False,
    num_classes=1000,
    activation="relu",
    classifier_activation="softmax",
    dropout=0,
    pretrained=None,
    model_name="vanillanet",
    kwargs=None,
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)

    inner_activation = "leaky_relu/{}".format(leaky_relu_alpha)
    stem_width = stem_width if stem_width > 0 else out_channels[0]
    nn = conv2d_no_bias(inputs, stem_width, kernel_size=4, strides=4, use_bias=True, name="stem_1_")
    if not deploy:
        nn = batchnorm_with_activation(nn, epsilon=BATCH_NORM_EPSILON, activation=inner_activation, name="stem_1_")

        nn = conv2d_no_bias(nn, stem_width, kernel_size=1, strides=1, use_bias=True, name="stem_2_")
        nn = batchnorm_with_activation(nn, epsilon=BATCH_NORM_EPSILON, activation=None, name="stem_2_")
    nn = activation_depthwise_conv_bn(nn, deploy=deploy, name="stem_2_act_")

    """ stages """
    for stack_id, (stride, out_channel) in enumerate(zip(strides, out_channels)):
        stack_name = "stack{}_".format(stack_id + 1)
        if not deploy:
            input_channel = nn.shape[-1 if image_data_format() == "channels_last" else 1]
            nn = conv2d_no_bias(nn, input_channel, kernel_size=1, use_bias=True, name=stack_name + "1_")
            nn = batchnorm_with_activation(nn, epsilon=BATCH_NORM_EPSILON, activation=inner_activation, name=stack_name + "1_")
        nn = conv2d_no_bias(nn, out_channel, kernel_size=1, use_bias=True, name=stack_name + ("1_" if deploy else "2_"))
        if not deploy:
            nn = batchnorm_with_activation(nn, epsilon=BATCH_NORM_EPSILON, activation=None, name=stack_name + "2_")

        if stride > 1:
            nn = layers.MaxPool2D(stride, name=stack_name + "down")(nn)
        nn = activation_depthwise_conv_bn(nn, deploy=deploy, name=stack_name + "act_")

    """ output """
    if num_classes > 0:
        nn = layers.GlobalAveragePooling2D(keepdims=True)(nn)  # tf.reduce_mean(nn, axis=1)
        if dropout > 0 and dropout < 1:
            nn = layers.Dropout(dropout)(nn)
        nn = conv2d_no_bias(nn, num_classes, kernel_size=1, use_bias=True, name="pre_head_")
        if not deploy:
            nn = batchnorm_with_activation(nn, epsilon=BATCH_NORM_EPSILON, activation=inner_activation, name="pre_head_")
            nn = conv2d_no_bias(nn, num_classes, kernel_size=1, use_bias=True, name="head_")
        if classifier_activation is not None:
            nn = activation_by_name(nn, activation=classifier_activation, name="head_")
        nn = layers.Flatten(dtype="float32", name="head")(nn)

    model = models.Model(inputs, nn, name=model_name)
    reload_model_weights(model, PRETRAINED_DICT, "vanillanet", pretrained)

    model.set_leaky_relu_alpha = lambda alpha: set_leaky_relu_alpha(model, alpha)
    model.switch_to_deploy = lambda: switch_to_deploy(model)
    add_pre_post_process(model, rescale_mode="torch")
    return model


def set_leaky_relu_alpha(model, alpha):
    for ii in model.layers:
        if isinstance(ii, layers.LeakyReLU):
            print("set alpha for {}, {} -> {}".format(ii.name, ii.alpha, alpha))
            ii.alpha = alpha


def switch_to_deploy(model):
    from keras_cv_attention_models import model_surgery

    new_model = model_surgery.convert_to_fused_conv_bn_model(model)
    remove_layer_condition = lambda layer: layer["class_name"] == "LeakyReLU" and layer["config"]["alpha"] == 1
    new_model = model_surgery.remove_layer_single_input(new_model, remove_layer_condition=remove_layer_condition)
    new_model = model_surgery.fuse_sequential_conv_strict(new_model)
    add_pre_post_process(new_model, rescale_mode=model.preprocess_input.rescale_mode, post_process=model.decode_predictions)
    return new_model


@register_model
def VanillaNet5(input_shape=(224, 224, 3), num_classes=1000, deploy=False, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return VanillaNet(**locals(), model_name="vanillanet_5" + ("_deploy" if deploy else ""), **kwargs)


@register_model
def VanillaNet6(input_shape=(224, 224, 3), num_classes=1000, deploy=False, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    out_channels = [1024, 2048, 4096, 4096]
    strides = [2, 2, 2, 1]
    return VanillaNet(**locals(), model_name="vanillanet_6" + ("_deploy" if deploy else ""), **kwargs)


@register_model
def VanillaNet7(input_shape=(224, 224, 3), num_classes=1000, deploy=False, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    out_channels = [512, 1024, 2048, 4096, 4096]
    strides = [1, 2, 2, 2, 1]
    return VanillaNet(**locals(), model_name="vanillanet_7" + ("_deploy" if deploy else ""), **kwargs)


@register_model
def VanillaNet8(input_shape=(224, 224, 3), num_classes=1000, deploy=False, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    out_channels = [512, 1024, 2048, 2048, 4096, 4096]
    strides = [1, 2, 2, 1, 2, 1]
    return VanillaNet(**locals(), model_name="vanillanet_8" + ("_deploy" if deploy else ""), **kwargs)


@register_model
def VanillaNet9(input_shape=(224, 224, 3), num_classes=1000, deploy=False, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    out_channels = [512, 1024] + [2048] * 3 + [4096, 4096]
    strides = [1, 2, 2] + [1] * 2 + [2, 1]
    return VanillaNet(**locals(), model_name="vanillanet_9" + ("_deploy" if deploy else ""), **kwargs)


@register_model
def VanillaNet10(
    input_shape=(224, 224, 3), num_classes=1000, deploy=False, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs
):
    out_channels = [512, 1024] + [2048] * 4 + [4096, 4096]
    strides = [1, 2, 2] + [1] * 3 + [2, 1]
    return VanillaNet(**locals(), model_name="vanillanet_10" + ("_deploy" if deploy else ""), **kwargs)


@register_model
def VanillaNet11(
    input_shape=(224, 224, 3), num_classes=1000, deploy=False, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs
):
    out_channels = [512, 1024] + [2048] * 5 + [4096, 4096]
    strides = [1, 2, 2] + [1] * 4 + [2, 1]
    return VanillaNet(**locals(), model_name="vanillanet_11" + ("_deploy" if deploy else ""), **kwargs)


@register_model
def VanillaNet12(
    input_shape=(224, 224, 3), num_classes=1000, deploy=False, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs
):
    out_channels = [512, 1024] + [2048] * 6 + [4096, 4096]
    strides = [1, 2, 2] + [1] * 5 + [2, 1]
    return VanillaNet(**locals(), model_name="vanillanet_12" + ("_deploy" if deploy else ""), **kwargs)


@register_model
def VanillaNet13(
    input_shape=(224, 224, 3), num_classes=1000, deploy=False, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs
):
    out_channels = [512, 1024] + [2048] * 7 + [4096, 4096]
    strides = [1, 2, 2] + [1] * 6 + [2, 1]
    return VanillaNet(**locals(), model_name="vanillanet_13" + ("_deploy" if deploy else ""), **kwargs)



from keras_cv_attention_models import backend as __backend__

from keras_cv_attention_models.clip.tokenizer import SimpleTokenizer, GPT2Tokenizer, TikToken, SentencePieceTokenizer
from keras_cv_attention_models.clip.models import (
    add_text_model_index_header,
    build_text_model_from_image_model,
    convert_to_clip_model,
    split_to_image_text_model,
    RunPrediction,
)
from keras_cv_attention_models.plot_func import plot_hists

if __backend__.is_tensorflow_backend:
    from keras_cv_attention_models.clip import tf_data as data
    from keras_cv_attention_models.clip.tf_data import init_dataset
    from keras_cv_attention_models.imagenet.data import show_batch_sample
else:
    from keras_cv_attention_models.clip import torch_data as data
    from keras_cv_attention_models.clip.torch_data import init_dataset



COCO_LABELS = """person, bicycle, car, motorcycle, airplane, bus, train, truck, boat, traffic light, fire hydrant, stop sign,
    parking meter, bench, bird, cat, dog, horse, sheep, cow, elephant, bear, zebra, giraffe, backpack, umbrella, handbag, tie,
    suitcase, frisbee, skis, snowboard, sports ball, kite, baseball bat, baseball glove, skateboard, surfboard, tennis racket,
    bottle, wine glass, cup, fork, knife, spoon, bowl, banana, apple, sandwich, orange, broccoli, carrot, hot dog, pizza, donut,
    cake, chair, couch, potted plant, bed, dining table, toilet, tv, laptop, mouse, remote, keyboard, cell phone, microwave, oven,
    toaster, sink, refrigerator, book, clock, vase, scissors, teddy bear, hair drier, toothbrush"""
COCO_80_LABEL_DICT = {id: ii.strip() for id, ii in enumerate(COCO_LABELS.split(","))}
INVALID_ID_90 = [11, 25, 28, 29, 44, 65, 67, 68, 70, 82]
COCO_90_LABEL_DICT = {id: ii for id, ii in zip(set(range(90)) - set(INVALID_ID_90), COCO_80_LABEL_DICT.values())}
COCO_90_LABEL_DICT.update({ii: "Unknown" for ii in INVALID_ID_90})
COCO_80_to_90_LABEL_DICT = {id_80: id_90 for id_80, id_90 in enumerate(set(range(90)) - set(INVALID_ID_90))}



from keras_cv_attention_models.mlp_family.mlp_mixer import (
    MLPMixer,
    MLPMixerS32,
    MLPMixerS16,
    MLPMixerB32,
    MLPMixerB16,
    MLPMixerL32,
    MLPMixerL16,
    MLPMixerH14,
    mlp_block,
    mlp_mixer_block,
)
from keras_cv_attention_models.mlp_family.res_mlp import ResMLP, ResMLP12, ResMLP24, ResMLP36, ResMLP_B24, ChannelAffine
from keras_cv_attention_models.mlp_family.gated_mlp import GMLP, GMLPTiny16, GMLPS16, GMLPB16, spatial_gating_block
from keras_cv_attention_models.mlp_family.wave_mlp import WaveMLP, WaveMLP_T, WaveMLP_S, WaveMLP_M, WaveMLP_B, phase_aware_token_mixing
from keras_cv_attention_models.mlp_family import mlp_mixer, res_mlp, gated_mlp, wave_mlp

__mlp_mixer_head_doc__ = """
Github source [leondgarse/keras_cv_attention_models](https://github.com/leondgarse/keras_cv_attention_models).
Keras implementation of [Github google-research/vision_transformer](https://github.com/google-research/vision_transformer#available-mixer-models).
Paper [PDF 2105.01601 MLP-Mixer: An all-MLP Architecture for Vision](https://arxiv.org/pdf/2105.01601.pdf).
"""

__tail_doc__ = """  input_shape: it should have exactly 3 inputs channels like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
      For `"imagenet21k"` pre-trained model, actual `num_classes` is `21843`.
  activation: activation used in whole model, default `gelu`.
  sam_rho: None zero value to init model using `SAM` training step.
      SAM Arxiv article: [Sharpness-Aware Minimization for Efficiently Improving Generalization](https://arxiv.org/pdf/2010.01412.pdf).
  dropout: top dropout rate if top layers is included.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be a constant value like `0.2`,
      or a tuple value like `(0, 0.2)` indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  classifier_activation: A `str` or callable. The activation function to use on the `top` layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the `top` layer.
      Default is `softmax`.
  pretrained: value in {pretrained_list}.
      Will try to download and load pre-trained model weights if not None.
      Save path is `~/.keras/models/`.

Returns:
    A `keras.Model` instance.
"""

MLPMixer.__doc__ = __mlp_mixer_head_doc__ + """
Args:
  num_blocks: number of layers.
  patch_size: stem patch resolution P×P, means `kernel_size=patch_size, strides=patch_size` for stem `Conv2D` block.
  stem_width: stem output channel dimenion.
  tokens_mlp_dim: MLP block token level hidden dimenion, where token level means `height * weight` dimension.
  channels_mlp_dim: MLP block channel level hidden dimenion.
  model_name: string, model name.
""" + __tail_doc__.format(pretrained_list=[None, "imagenet", "imagenet21k", "imagenet_sam"]) + """
Model architectures:
  | Model            | Params | FLOPs   | Input | Top1 Acc |
  | ---------------- | ------ | ------- | ----- | -------- |
  | MLPMixerS32, JFT | 19.1M  | 1.01G   | 224   | 68.70    |
  | MLPMixerS16, JFT | 18.5M  | 3.79G   | 224   | 73.83    |
  | MLPMixerB32, JFT | 60.3M  | 3.25G   | 224   | 75.53    |
  | - imagenet_sam   | 60.3M  | 3.25G   | 224   | 72.47    |
  | MLPMixerB16      | 59.9M  | 12.64G  | 224   | 76.44    |
  | - imagenet21k    | 59.9M  | 12.64G  | 224   | 80.64    |
  | - imagenet_sam   | 59.9M  | 12.64G  | 224   | 77.36    |
  | - JFT            | 59.9M  | 12.64G  | 224   | 80.00    |
  | MLPMixerL32, JFT | 206.9M | 11.30G  | 224   | 80.67    |
  | MLPMixerL16      | 208.2M | 44.66G  | 224   | 71.76    |
  | - imagenet21k    | 208.2M | 44.66G  | 224   | 82.89    |
  | - input 448      | 208.2M | 178.54G | 448   | 83.91    |
  | - input 224, JFT | 208.2M | 44.66G  | 224   | 84.82    |
  | - input 448, JFT | 208.2M | 178.54G | 448   | 86.78    |
  | MLPMixerH14, JFT | 432.3M | 121.22G | 224   | 86.32    |
  | - input 448, JFT | 432.3M | 484.73G | 448   | 87.94    |

  | Specification        | S/32  | S/16  | B/32  | B/16  | L/32  | L/16  | H/14  |
  | -------------------- | ----- | ----- | ----- | ----- | ----- | ----- | ----- |
  | Number of layers     | 8     | 8     | 12    | 12    | 24    | 24    | 32    |
  | Patch resolution P×P | 32×32 | 16×16 | 32×32 | 16×16 | 32×32 | 16×16 | 14×14 |
  | Hidden size C        | 512   | 512   | 768   | 768   | 1024  | 1024  | 1280  |
  | Sequence length S    | 49    | 196   | 49    | 196   | 49    | 196   | 256   |
  | MLP dimension DC     | 2048  | 2048  | 3072  | 3072  | 4096  | 4096  | 5120  |
  | MLP dimension DS     | 256   | 256   | 384   | 384   | 512   | 512   | 640   |
"""


__mixer_default_doc__ = __mlp_mixer_head_doc__ + """
[{model_name} architecture] num_blocks: {num_blocks}, patch_size: {patch_size}, stem_width: {stem_width}, tokens_mlp_dim: {tokens_mlp_dim}, channels_mlp_dim: {channels_mlp_dim}.

Args:
""" + __tail_doc__.format(pretrained_list=[None, "imagenet", "imagenet21k", "imagenet_sam"])

MLPMixerS32.__doc__ = __mixer_default_doc__.format(model_name="MLPMixerS32", **mlp_mixer.BLOCK_CONFIGS["s32"])
MLPMixerS16.__doc__ = __mixer_default_doc__.format(model_name="MLPMixerS16", **mlp_mixer.BLOCK_CONFIGS["s16"])
MLPMixerB32.__doc__ = __mixer_default_doc__.format(model_name="MLPMixerB32", **mlp_mixer.BLOCK_CONFIGS["b32"])
MLPMixerB16.__doc__ = __mixer_default_doc__.format(model_name="MLPMixerB16", **mlp_mixer.BLOCK_CONFIGS["b16"])
MLPMixerL32.__doc__ = __mixer_default_doc__.format(model_name="MLPMixerL32", **mlp_mixer.BLOCK_CONFIGS["l32"])
MLPMixerL16.__doc__ = __mixer_default_doc__.format(model_name="MLPMixerL16", **mlp_mixer.BLOCK_CONFIGS["l16"])
MLPMixerH14.__doc__ = __mixer_default_doc__.format(model_name="MLPMixerH14", **mlp_mixer.BLOCK_CONFIGS["h14"])

mlp_block.__doc__ = __mlp_mixer_head_doc__ + """
MLP block

Args:
  inputs: Input tensor.
  hidden_dim: Expanded channel dimension for the first `Dense` layer.
  output_channel: Number of block output channels. Default `-1` for same with inputs.
  drop_rate: Dropout rate for inner and output. Default `0` for not using Dropout.
  use_conv: Boolean value whether using `Dense` or `Conv2D` layer. Default False.
  activation: activation applied, default `gelu`.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> inputs = keras.layers.Input([14 * 14, 9])
>>> nn = attention_layers.mlp_block(inputs, 9 * 4)
>>> print(f"{nn.shape = }")
nn.shape = TensorShape([None, 196, 9])

>>> keras.models.Model(inputs, nn).summary()
_________________________________________________________________
Layer (type)                 Output Shape              Param #
=================================================================
input_1 (InputLayer)         [(None, 196, 9)]          0
_________________________________________________________________
dense (Dense)                (None, 196, 36)           360
_________________________________________________________________
activation (Activation)      (None, 196, 36)           0
_________________________________________________________________
dense_1 (Dense)              (None, 196, 9)            333
=================================================================
Total params: 693
Trainable params: 693
Non-trainable params: 0
_________________________________________________________________
"""
mlp_mixer_block.__doc__ = __mlp_mixer_head_doc__ + """
MLP Mixer block

Args:
  inputs: Input tensor.
  tokens_mlp_dim: hidden_dim for the first mlp_block.
  channels_mlp_dim: hidden_dim for the second mlp_block.
  drop_rate: block dropout rate.
  activation: activation for mlp_block.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> inputs = keras.layers.Input([14 * 14, 9])
>>> nn = attention_layers.mlp_mixer_block(inputs, tokens_mlp_dim=14 * 14, channels_mlp_dim=9 * 4)
>>> print(f"{nn.shape = }")
nn.shape = TensorShape([None, 196, 9])

>>> mm = keras.models.Model(inputs, nn)
>>> mm.summary()
__________________________________________________________________________________________________
Layer (type)                    Output Shape         Param #     Connected to
==================================================================================================
input_1 (InputLayer)            [(None, 196, 9)]     0
__________________________________________________________________________________________________
layer_normalization (LayerNorma (None, 196, 9)       18          input_1[0][0]
__________________________________________________________________________________________________
permute (Permute)               (None, 9, 196)       0           layer_normalization[0][0]
__________________________________________________________________________________________________
dense (Dense)                   (None, 9, 196)       38612       permute[0][0]
__________________________________________________________________________________________________
activation (Activation)         (None, 9, 196)       0           dense[0][0]
__________________________________________________________________________________________________
dense_1 (Dense)                 (None, 9, 196)       38612       activation[0][0]
__________________________________________________________________________________________________
permute_1 (Permute)             (None, 196, 9)       0           dense_1[0][0]
__________________________________________________________________________________________________
add (Add)                       (None, 196, 9)       0           permute_1[0][0]
                                                                 input_1[0][0]
__________________________________________________________________________________________________
layer_normalization_1 (LayerNor (None, 196, 9)       18          add[0][0]
__________________________________________________________________________________________________
dense_2 (Dense)                 (None, 196, 36)      360         layer_normalization_1[0][0]
__________________________________________________________________________________________________
activation_1 (Activation)       (None, 196, 36)      0           dense_2[0][0]
__________________________________________________________________________________________________
dense_3 (Dense)                 (None, 196, 9)       333         activation_1[0][0]
__________________________________________________________________________________________________
add_1 (Add)                     (None, 196, 9)       0           dense_3[0][0]
                                                                 add[0][0]
==================================================================================================
Total params: 77,953
Trainable params: 77,953
Non-trainable params: 0
__________________________________________________________________________________________________
"""

__resmlp_head_doc__ = """
Github source [leondgarse/keras_cv_attention_models](https://github.com/leondgarse/keras_cv_attention_models).
Keras implementation of [Github facebookresearch/resmlp_models.py](https://github.com/facebookresearch/deit/blob/main/resmlp_models.py).
Paper [PDF 2105.03404 ResMLP: Feedforward networks for image classification with data-efficient training](https://arxiv.org/pdf/2105.03404.pdf).
"""

ResMLP.__doc__ = __resmlp_head_doc__ + """
Args:
  num_blocks: number of layers.
  patch_size: stem patch resolution P×P, means `kernel_size=patch_size, strides=patch_size` for stem `Conv2D` block.
  stem_width: stem output channel dimenion.
  channels_mlp_dim: MLP block channel level hidden dimenion.
  model_name: string, model name.
""" + __tail_doc__.format(pretrained_list=[None, "imagenet", "imagenet22k"]) + """
Model architectures:
  | Model         | Params | FLOPs   | Input | Top1 Acc |
  | ------------- | ------ | ------- | ----- | -------- |
  | ResMLP12      | 15M    | 3.02G   | 224   | 77.8     |
  | ResMLP24      | 30M    | 5.98G   | 224   | 80.8     |
  | ResMLP36      | 116M   | 8.94G   | 224   | 81.1     |
  | ResMLP_B24    | 129M   | 100.39G | 224   | 83.6     |
  | - imagenet22k | 129M   | 100.39G | 224   | 84.4     |
"""

__resmlp_default_doc__ = __resmlp_head_doc__ + """
[{model_name} architecture] num_blocks: {num_blocks}, patch_size: {patch_size}, stem_width: {stem_width}, channels_mlp_dim: {channels_mlp_dim}.

Args:
""" + __tail_doc__.format(pretrained_list=[None, "imagenet", "imagenet22k"])

ResMLP12.__doc__ = __resmlp_default_doc__.format(model_name="ResMLP12", **res_mlp.BLOCK_CONFIGS["12"])
ResMLP24.__doc__ = __resmlp_default_doc__.format(model_name="ResMLP24", **res_mlp.BLOCK_CONFIGS["24"])
ResMLP36.__doc__ = __resmlp_default_doc__.format(model_name="ResMLP36", **res_mlp.BLOCK_CONFIGS["36"])
ResMLP_B24.__doc__ = __resmlp_default_doc__.format(model_name="ResMLP_B24", **res_mlp.BLOCK_CONFIGS["b24"])

ChannelAffine.__doc__ = __resmlp_head_doc__ + """
ChannelAffine layer

Examples:
>>> from keras_cv_attention_models import attention_layers
>>> aa = attention_layers.ChannelAffine()
>>> print(f"{aa(tf.ones([1, 32, 32, 192])).shape = }")
aa(tf.ones([1, 32, 32, 192])).shape = TensorShape([1, 32, 32, 192])

>>> print({ii.name: ii.shape for ii in aa.weights})
{'channel_affine_1/weight:0': TensorShape([192]),
 'channel_affine_1/bias:0': TensorShape([192])}
"""

__gmlp_head_doc__ = """
Github source [leondgarse/keras_cv_attention_models](https://github.com/leondgarse/keras_cv_attention_models).
Keras implementation of Gated MLP.
Paper [PDF 2105.08050 Pay Attention to MLPs](https://arxiv.org/pdf/2105.08050.pdf).
"""

GMLP.__doc__ = __gmlp_head_doc__ + """
Args:
  num_blocks: number of layers.
  patch_size: stem patch resolution P×P, means `kernel_size=patch_size, strides=patch_size` for stem `Conv2D` block.
  stem_width: stem output channel dimenion.
  channels_mlp_dim: MLP block channel level hidden dimenion.
  model_name: string, model name.
""" + __tail_doc__.format(pretrained_list=[None, "imagenet"]) + """
Model architectures:
  | Model      | Params | FLOPs  | Input | Top1 Acc |
  | ---------- | ------ | ------ | ----- | -------- |
  | GMLPTiny16 | 6M     | 1.35G  | 224   | 72.3     |
  | GMLPS16    | 20M    | 4.44G  | 224   | 79.6     |
  | GMLPB16    | 73M    | 15.82G | 224   | 81.6     |
"""

__gmlp_default_doc__ = __gmlp_head_doc__ + """
[{model_name} architecture] num_blocks: {num_blocks}, patch_size: {patch_size}, stem_width: {stem_width}, channels_mlp_dim: {channels_mlp_dim}.

Args:
""" + __tail_doc__.format(pretrained_list=[None, "imagenet"])

GMLPTiny16.__doc__ = __gmlp_default_doc__.format(model_name="GMLPTiny16", **gated_mlp.BLOCK_CONFIGS["tiny16"])
GMLPS16.__doc__ = __gmlp_default_doc__.format(model_name="GMLPS16", **gated_mlp.BLOCK_CONFIGS["s16"])
GMLPB16.__doc__ = __gmlp_default_doc__.format(model_name="GMLPB16", **gated_mlp.BLOCK_CONFIGS["b16"])

spatial_gating_block.__doc__ = __gmlp_head_doc__ + """
Spatial Gating Block

input: `[batch, height * width, channel]`.
output: `[batch, height * width, channel // 2]`.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> inputs = keras.layers.Input([14 * 14, 8])
>>> nn = attention_layers.spatial_gating_block(inputs)
>>> mm = keras.models.Model(inputs, nn)
>>> mm.summary()
__________________________________________________________________________________________________
Layer (type)                    Output Shape         Param #     Connected to
==================================================================================================
input_1 (InputLayer)            [(None, 196, 8)]     0
__________________________________________________________________________________________________
tf.split (TFOpLambda)           [(None, 196, 4), (No 0           input_1[0][0]
__________________________________________________________________________________________________
layer_normalization (LayerNorma (None, 196, 4)       8           tf.split[0][1]
__________________________________________________________________________________________________
permute (Permute)               (None, 4, 196)       0           layer_normalization[0][0]
__________________________________________________________________________________________________
dense (Dense)                   (None, 4, 196)       38612       permute[0][0]
__________________________________________________________________________________________________
permute_1 (Permute)             (None, 196, 4)       0           dense[0][0]
__________________________________________________________________________________________________
multiply (Multiply)             (None, 196, 4)       0           tf.split[0][0]
                                                                 permute_1[0][0]
==================================================================================================
Total params: 38,620
Trainable params: 38,620
Non-trainable params: 0
__________________________________________________________________________________________________
"""

__wavemlp_head_doc__ = """
Github source [leondgarse/keras_cv_attention_models](https://github.com/leondgarse/keras_cv_attention_models).
Keras implementation of [Github huawei-noah/wavemlp_pytorch](https://github.com/huawei-noah/CV-Backbones/tree/master/wavemlp_pytorch).
Paper [PDF 2111.12294 An Image Patch is a Wave: Quantum Inspired Vision MLP](https://arxiv.org/pdf/2111.12294.pdf).
"""

WaveMLP.__doc__ = __wavemlp_head_doc__ + """
Args:
  num_blocks: number of layers.
  out_channels: output channel for each block.
  stem_width: stem output channel dimenion. Default -1 for out_channels[0]
  mlp_ratios: mlp hidden channel expansion ratios for each block.
  use_downsample_norm: boolean value if stem and down saple blocks using norm layers.
      True for WaveMLP_T and WaveMLP_S, False for WaveMLP_M and WaveMLP_B.
  use_group_norm: boolean value if using GroupNormalization instead of BatchNormalization. False for WaveMLP_T, True for others.
  qkv_bias: boolean value if use bias for `phase_aware_token_mixing` blocks.
  model_name: string, model name.
""" + __tail_doc__.format(pretrained_list=[None, "imagenet"]) + """
Model architectures:
  | Model     | Params | FLOPs  | Input | Top1 Acc |
  | --------- | ------ | ------ | ----- | -------- |
  | WaveMLP_T | 17M    | 2.47G  | 224   | 80.9     |
  | WaveMLP_S | 30M    | 4.55G  | 224   | 82.9     |
  | WaveMLP_M | 44M    | 7.92G  | 224   | 83.3     |
  | WaveMLP_B | 63M    | 10.26G | 224   | 83.6     |
"""

WaveMLP_T.__doc__ = __wavemlp_head_doc__ + """
Args:
""" + __tail_doc__.format(pretrained_list=[None, "imagenet"])

WaveMLP_S.__doc__ = WaveMLP_T.__doc__
WaveMLP_M.__doc__ = WaveMLP_T.__doc__
WaveMLP_B.__doc__ = WaveMLP_T.__doc__

phase_aware_token_mixing.__doc__ = __wavemlp_head_doc__ + """
Phase-Aware Token Mixing Block

input: `[batch, height, width, channel]`.
output: `[batch, height, width, out_channel]`.

Args:
  inputs: input tensor.
  out_channel: output channel, default -1 for same with inputs.
  qkv_bias: boolean value if use bias for `height` / `width` / `channel` conv layers, default False.
  output_dropout: dropout rate for output, default 0.
  activation: activation for internal mlp block, default "gelu".

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> inputs = keras.layers.Input([14, 14, 8])
>>> nn = attention_layers.phase_aware_token_mixing(inputs, name="aa")
>>> mm = keras.models.Model(inputs, nn)
>>> mm.summary()
# Total params: 786
# Trainable params: 754
# Non-trainable params: 32

>>> print({ii.name: [list(jj.shape) for jj in ii.weights] for ii in mm.layers if len(ii.weights) != 0})
# {'aa_theta_h_conv': [[1, 1, 8, 8], [8]],
#  'aa_theta_w_conv': [[1, 1, 8, 8], [8]],
#  'aa_theta_h_bn': [[8], [8], [8], [8]],
#  'aa_theta_w_bn': [[8], [8], [8], [8]],
#  'aa_height_conv': [[1, 1, 8, 8]],
#  'aa_width_conv': [[1, 1, 8, 8]],
#  'aa_height_down_conv': [[1, 7, 2, 8]],
#  'aa_width_down_conv': [[7, 1, 2, 8]],
#  'aa_channel_conv': [[1, 1, 8, 8]],
#  'aa_reweight_Conv_0': [[1, 1, 8, 2], [2]],
#  'aa_reweight_Conv_1': [[1, 1, 2, 24], [24]],
#  'aa_out_conv': [[1, 1, 8, 8], [8]]}
"""



from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import activation_by_name, add_pre_post_process
from keras_cv_attention_models.download_and_load import reload_model_weights

BATCH_NORM_EPSILON = 1e-5

PRETRAINED_DICT = {
    "mlp_mixer_b16": {
        "imagenet21k": "6353dffc590a2a7348a44cee2c784724",
        "imagenet": "abd04090063ba9ab0d49e2131cef9d64",
        "imagenet_sam": "d953ef41ffdb0ab9c3fa21493bf0982f",
    },
    "mlp_mixer_l16": {"imagenet": "fa91a74f1aa11ed610299d06d643ed45", "imagenet21k": "8dca5de1817112d9e717db6b2e9a7b0b"},
    "mlp_mixer_b32": {"imagenet_sam": "a6285750e55579fc68e7ba68a683c77d"},
}


def layer_norm(inputs, axis="auto", name=None):
    """Typical LayerNormalization with epsilon=1e-5"""
    norm_axis = (-1 if backend.image_data_format() == "channels_last" else 1) if axis == "auto" else axis
    return layers.LayerNormalization(axis=norm_axis, epsilon=BATCH_NORM_EPSILON, name=name)(inputs)


def mlp_block(inputs, hidden_dim, output_channel=-1, drop_rate=0, use_conv=False, use_bias=True, activation="gelu", name=None):
    channnel_axis = -1 if image_data_format() == "channels_last" else (1 if use_conv else -1)
    output_channel = output_channel if output_channel > 0 else inputs.shape[channnel_axis]
    if use_conv:
        nn = layers.Conv2D(hidden_dim, kernel_size=1, use_bias=use_bias, name=name and name + "Conv_0")(inputs)
    else:
        nn = layers.Dense(hidden_dim, use_bias=use_bias, name=name and name + "Dense_0")(inputs)
    nn = activation_by_name(nn, activation, name=name)
    nn = layers.Dropout(drop_rate)(nn) if drop_rate > 0 else nn
    if use_conv:
        nn = layers.Conv2D(output_channel, kernel_size=1, use_bias=use_bias, name=name and name + "Conv_1")(nn)
    else:
        nn = layers.Dense(output_channel, use_bias=use_bias, name=name and name + "Dense_1")(nn)
    nn = layers.Dropout(drop_rate)(nn) if drop_rate > 0 else nn
    return nn


def mlp_mixer_block(inputs, tokens_mlp_dim, channels_mlp_dim, use_bias=True, drop_rate=0, activation="gelu", data_format=None, name=None):
    data_format = backend.image_data_format() if data_format is None else data_format
    norm_axis = -1 if data_format == "channels_last" else 1

    nn = layer_norm(inputs, axis=norm_axis, name=name and name + "LayerNorm_0")
    nn = layers.Permute((2, 1), name=name and name + "permute_0")(nn) if data_format == "channels_last" else nn
    nn = mlp_block(nn, tokens_mlp_dim, use_bias=use_bias, activation=activation, name=name and name + "token_mixing/")
    nn = layers.Permute((2, 1), name=name and name + "permute_1")(nn) if data_format == "channels_last" else nn
    if drop_rate > 0:
        nn = layers.Dropout(drop_rate, noise_shape=(None, 1, 1), name=name and name + "token_drop")(nn)
    token_out = layers.Add(name=name and name + "add_0")([nn, inputs])

    nn = layer_norm(token_out, axis=norm_axis, name=name and name + "LayerNorm_1")
    nn = nn if data_format == "channels_last" else layers.Permute((2, 1), name=name and name + "permute_2")(nn)
    nn = mlp_block(nn, channels_mlp_dim, use_bias=use_bias, activation=activation, name=name and name + "channel_mixing/")
    channel_out = nn if data_format == "channels_last" else layers.Permute((2, 1), name=name and name + "permute_3")(nn)
    if drop_rate > 0:
        channel_out = layers.Dropout(drop_rate, noise_shape=(None, 1, 1), name=name and name + "channel_drop")(channel_out)
    return layers.Add(name=name and name + "output")([channel_out, token_out])


def MLPMixer(
    num_blocks,
    patch_size,
    stem_width,
    tokens_mlp_dim,
    channels_mlp_dim,
    input_shape=(224, 224, 3),
    num_classes=0,
    activation="gelu",
    sam_rho=0,
    dropout=0,
    drop_connect_rate=0,
    classifier_activation="softmax",
    pretrained="imagenet",
    model_name="mlp_mixer",
    kwargs=None,
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)

    nn = layers.Conv2D(stem_width, kernel_size=patch_size, strides=patch_size, padding="valid", name="stem")(inputs)
    new_shape = [nn.shape[1] * nn.shape[2], stem_width] if backend.image_data_format() == "channels_last" else [stem_width, nn.shape[2] * nn.shape[3]]
    nn = layers.Reshape(new_shape)(nn)

    drop_connect_s, drop_connect_e = drop_connect_rate if isinstance(drop_connect_rate, (list, tuple)) else [drop_connect_rate, drop_connect_rate]
    for ii in range(num_blocks):
        name = "{}_{}/".format("MixerBlock", str(ii))
        block_drop_rate = drop_connect_s + (drop_connect_e - drop_connect_s) * ii / num_blocks
        nn = mlp_mixer_block(nn, tokens_mlp_dim, channels_mlp_dim, drop_rate=block_drop_rate, activation=activation, name=name)
    nn = layer_norm(nn, name="pre_head_layer_norm")

    if num_classes > 0:
        nn = layers.GlobalAveragePooling1D()(nn)  # tf.reduce_mean(nn, axis=1)
        if dropout > 0 and dropout < 1:
            nn = layers.Dropout(dropout)(nn)
        nn = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="head")(nn)

    if sam_rho != 0:
        from keras_cv_attention_models.model_surgery import SAMModel

        model = SAMModel(inputs, nn, name=model_name)
    else:
        model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="tf")
    reload_model_weights(model, pretrained_dict=PRETRAINED_DICT, sub_release="mlp_family", pretrained=pretrained)
    return model


BLOCK_CONFIGS = {
    "s32": {
        "num_blocks": 8,
        "patch_size": 32,
        "stem_width": 512,
        "tokens_mlp_dim": 256,
        "channels_mlp_dim": 2048,
    },
    "s16": {
        "num_blocks": 8,
        "patch_size": 16,
        "stem_width": 512,
        "tokens_mlp_dim": 256,
        "channels_mlp_dim": 2048,
    },
    "b32": {
        "num_blocks": 12,
        "patch_size": 32,
        "stem_width": 768,
        "tokens_mlp_dim": 384,
        "channels_mlp_dim": 3072,
    },
    "b16": {
        "num_blocks": 12,
        "patch_size": 16,
        "stem_width": 768,
        "tokens_mlp_dim": 384,
        "channels_mlp_dim": 3072,
    },
    "l32": {
        "num_blocks": 24,
        "patch_size": 32,
        "stem_width": 1024,
        "tokens_mlp_dim": 512,
        "channels_mlp_dim": 4096,
    },
    "l16": {
        "num_blocks": 24,
        "patch_size": 16,
        "stem_width": 1024,
        "tokens_mlp_dim": 512,
        "channels_mlp_dim": 4096,
    },
    "h14": {
        "num_blocks": 32,
        "patch_size": 14,
        "stem_width": 1280,
        "tokens_mlp_dim": 640,
        "channels_mlp_dim": 5120,
    },
}


@register_model
def MLPMixerS32(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return MLPMixer(**BLOCK_CONFIGS["s32"], **locals(), model_name="mlp_mixer_s32", **kwargs)


@register_model
def MLPMixerS16(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return MLPMixer(**BLOCK_CONFIGS["s16"], **locals(), model_name="mlp_mixer_s16", **kwargs)


@register_model
def MLPMixerB32(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return MLPMixer(**BLOCK_CONFIGS["b32"], **locals(), model_name="mlp_mixer_b32", **kwargs)


@register_model
def MLPMixerB16(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return MLPMixer(**BLOCK_CONFIGS["b16"], **locals(), model_name="mlp_mixer_b16", **kwargs)


@register_model
def MLPMixerL32(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return MLPMixer(**BLOCK_CONFIGS["l32"], **locals(), model_name="mlp_mixer_l32", **kwargs)


@register_model
def MLPMixerL16(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return MLPMixer(**BLOCK_CONFIGS["l16"], **locals(), model_name="mlp_mixer_l16", **kwargs)


@register_model
def MLPMixerH14(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return MLPMixer(**BLOCK_CONFIGS["h14"], **locals(), model_name="mlp_mixer_h14", **kwargs)


if __name__ == "__convert__":
    aa = np.load("../models/imagenet1k_Mixer-B_16.npz")
    bb = {kk: vv for kk, vv in aa.items()}
    # cc = {kk: vv.shape for kk, vv in bb.items()}

    import mlp_mixer

    mm = mlp_mixer.MLPMixerB16(num_classes=1000, pretrained=None)
    # dd = {ii.name: ii.shape for ii in mm.weights}

    target_weights_dict = {"kernel": 0, "bias": 1, "scale": 0, "running_var": 3}
    for kk, vv in bb.items():
        split_name = kk.split("/")
        source_name = "/".join(split_name[:-1])
        source_weight_type = split_name[-1]
        target_layer = mm.get_layer(source_name)

        target_weights = target_layer.get_weights()
        target_weight_pos = target_weights_dict[source_weight_type]
        print("[{}] source: {}, target: {}".format(kk, vv.shape, target_weights[target_weight_pos].shape))

        target_weights[target_weight_pos] = vv
        target_layer.set_weights(target_weights)



from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    batchnorm_with_activation,
    conv2d_no_bias,
    depthwise_conv2d_no_bias,
    drop_block,
    layer_norm,
    se_module,
    output_block,
    MultiHeadRelativePositionalEmbedding,
    qkv_to_multi_head_channels_last_format,
    scaled_dot_product_attention,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

PRETRAINED_DICT = {"coatnet0": {"imagenet": {160: "bc4375d2f03b99ac4252770331f0d22f", 224: "5d1e563f959f7efb6d395bde7373ed26"}}}


def mhsa_with_multi_head_relative_position_embedding(
    inputs,
    num_heads=4,
    key_dim=0,
    global_query=None,
    out_shape=None,
    out_weight=True,
    qkv_bias=False,
    out_bias=False,
    attn_dropout=0,
    data_format=None,
    name=None,
):
    data_format = image_data_format() if data_format is None else data_format
    input_channel = inputs.shape[-1 if data_format == "channels_last" else 1]
    height, width = inputs.shape[1:-1] if data_format == "channels_last" else inputs.shape[2:]

    key_dim = key_dim if key_dim > 0 else input_channel // num_heads
    out_shape = input_channel if out_shape is None or not out_weight else out_shape
    qk_out = num_heads * key_dim
    # vv_dim = out_shape // num_heads
    vv_dim = key_dim
    blocks = height * width

    # Permute for conv if given data_format not matching actual image_data_format
    if image_data_format() == "channels_last" and data_format == "channels_first":
        inputs = layers.Permute([2, 3, 1])(inputs)
    elif image_data_format() == "channels_first" and data_format == "channels_last":
        inputs = layers.Permute([3, 1, 2])(inputs)
    conv_channel_axis = -1 if image_data_format() == "channels_last" else 1

    if global_query is not None:
        # kv = layers.Dense(qk_out * 2, use_bias=qkv_bias, name=name and name + "kv")(inputs)  # For GCViT weights
        kv = conv2d_no_bias(inputs, qk_out * 2, kernel_size=1, use_bias=qkv_bias, name=name and name + "kv_")
        kv = functional.reshape(kv, [-1, blocks, kv.shape[-1]] if image_data_format() == "channels_last" else [-1, kv.shape[1], blocks])
        key, value = functional.split(kv, [qk_out, out_shape], axis=conv_channel_axis)
        query = global_query
        _, key, value = qkv_to_multi_head_channels_last_format(None, key, value, num_heads=num_heads, data_format=None)
    else:
        # qkv = conv2d_no_bias(inputs, qk_out * 2 + out_shape, use_bias=qkv_bias, kernel_size=1, name=name and name + "qkv_")
        # qkv = layers.Dense(qk_out * 3, use_bias=qkv_bias, name=name and name + "qkv")(inputs)  # For GCViT weights
        # query = layers.Dense(qk_out, use_bias=qkv_bias, name=name and name + "query")(inputs)  # For MaxViT weights
        # key = layers.Dense(qk_out, use_bias=qkv_bias, name=name and name + "key")(inputs)  # For MaxViT weights
        # value = layers.Dense(qk_out, use_bias=qkv_bias, name=name and name + "value")(inputs)  # For MaxViT weights
        qkv = conv2d_no_bias(inputs, qk_out * 3, kernel_size=1, use_bias=qkv_bias, name=name and name + "qkv_")
        query, key, value = functional.split(qkv, [qk_out, qk_out, qk_out], axis=conv_channel_axis)
        query, key, value = qkv_to_multi_head_channels_last_format(query, key, value, num_heads=num_heads, data_format=None)
    # print(f"{query.shape = }, {key.shape = }, {value.shape = }, {num_heads = }, {key_dim = }, {vv_dim = }")

    pos_emb = MultiHeadRelativePositionalEmbedding(with_cls_token=False, attn_height=height, name=name and name + "pos_emb")
    output_shape = (height, width, out_shape)
    out = scaled_dot_product_attention(query, key, value, output_shape, pos_emb, out_weight=out_weight, out_bias=out_bias, dropout=attn_dropout, name=name)
    return out if data_format == "channels_last" else layers.Permute([3, 1, 2], name=name and name + "output_perm")(out)


def res_MBConv(
    inputs,
    output_channel,
    conv_short_cut=True,
    strides=1,
    expansion=4,
    se_ratio=0,
    drop_rate=0,
    use_dw_strides=True,
    bn_act_first=False,
    activation="gelu",
    name="",
):
    """x ← Proj(Pool(x)) + Conv (DepthConv (Conv (Norm(x), stride = 2))))"""
    preact = batchnorm_with_activation(inputs, activation=None, zero_gamma=False, name=name + "preact_")

    if conv_short_cut:
        shortcut = layers.MaxPool2D(strides, strides=strides, padding="same", name=name + "shortcut_pool")(inputs) if strides > 1 else inputs
        shortcut = conv2d_no_bias(shortcut, output_channel, 1, strides=1, name=name + "shortcut_")
        # shortcut = batchnorm_with_activation(shortcut, activation=activation, zero_gamma=False, name=name + "shortcut_")
    else:
        shortcut = inputs

    # MBConv
    input_channel = inputs.shape[-1 if image_data_format() == "channels_last" else 1]
    conv_strides, dw_strides = (1, strides) if use_dw_strides else (strides, 1)  # May swap stirdes with DW
    nn = conv2d_no_bias(preact, input_channel * expansion, 1, strides=conv_strides, use_bias=bn_act_first, padding="same", name=name + "expand_")
    nn = batchnorm_with_activation(nn, activation=activation, act_first=bn_act_first, name=name + "expand_")
    nn = depthwise_conv2d_no_bias(nn, 3, strides=dw_strides, use_bias=bn_act_first, padding="same", name=name + "MB_")
    nn = batchnorm_with_activation(nn, activation=activation, act_first=bn_act_first, zero_gamma=False, name=name + "MB_dw_")
    if se_ratio:
        nn = se_module(nn, se_ratio=se_ratio / expansion, activation=activation, name=name + "se_")
    nn = conv2d_no_bias(nn, output_channel, 1, strides=1, padding="same", name=name + "MB_pw_")
    # nn = batchnorm_with_activation(nn, activation=None, zero_gamma=True, name=name + "MB_pw_")
    nn = drop_block(nn, drop_rate=drop_rate, name=name)
    return layers.Add(name=name + "output")([shortcut, nn])


def res_ffn(inputs, expansion=4, kernel_size=1, drop_rate=0, activation="gelu", name=""):
    """x ← x + Module (Norm(x)), similar with typical MLP block"""
    # preact = batchnorm_with_activation(inputs, activation=None, zero_gamma=False, name=name + "preact_")
    preact = layer_norm(inputs, name=name + "preact_")

    input_channel = inputs.shape[-1 if image_data_format() == "channels_last" else 1]
    nn = conv2d_no_bias(preact, input_channel * expansion, kernel_size, name=name + "1_")
    nn = activation_by_name(nn, activation=activation, name=name)
    nn = conv2d_no_bias(nn, input_channel, kernel_size, name=name + "2_")
    nn = drop_block(nn, drop_rate=drop_rate, name=name)
    # return layers.Add(name=name + "output")([preact, nn])
    return layers.Add(name=name + "output")([inputs, nn])


def res_mhsa(inputs, output_channel, conv_short_cut=True, strides=1, head_dimension=32, drop_rate=0, activation="gelu", name=""):
    """x ← Proj(Pool(x)) + Attention (Pool(Norm(x)))"""
    # preact = batchnorm_with_activation(inputs, activation=None, zero_gamma=False, name=name + "preact_")
    preact = layer_norm(inputs, name=name + "preact_")

    if conv_short_cut:
        shortcut = layers.MaxPool2D(strides, strides=strides, padding="same", name=name + "shortcut_pool")(inputs) if strides > 1 else inputs
        shortcut = conv2d_no_bias(shortcut, output_channel, 1, strides=1, name=name + "shortcut_")
        # shortcut = batchnorm_with_activation(shortcut, activation=activation, zero_gamma=False, name=name + "shortcut_")
    else:
        shortcut = inputs

    nn = preact
    if strides != 1:  # Downsample
        # nn = layers.ZeroPadding2D(padding=1, name=name + "pad")(nn)
        nn = layers.MaxPool2D(pool_size=2, strides=strides, padding="same", name=name + "pool")(nn)
    num_heads = nn.shape[-1 if image_data_format() == "channels_last" else 1] // head_dimension
    nn = mhsa_with_multi_head_relative_position_embedding(nn, num_heads=num_heads, key_dim=head_dimension, out_shape=output_channel, name=name + "mhsa_")
    nn = drop_block(nn, drop_rate=drop_rate, name=name)
    # print(f"{name = }, {inputs.shape = }, {shortcut.shape = }, {nn.shape = }")
    return layers.Add(name=name + "output")([shortcut, nn])


def CoAtNet(
    num_blocks,
    out_channels,
    stem_width=64,
    block_types=["conv", "conv", "transform", "transform"],
    strides=[2, 2, 2, 2],
    expansion=4,
    se_ratio=0.25,
    head_dimension=32,
    use_dw_strides=True,
    bn_act_first=False,  # Experiment, use activation -> BatchNorm instead of BatchNorm -> activation, also set use_bias=True for pre Conv2D layer
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="gelu",
    drop_connect_rate=0,
    classifier_activation="softmax",
    dropout=0,
    pretrained=None,
    model_name="coatnet",
    kwargs=None,
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)

    """ stage 0, Stem_stage """
    nn = conv2d_no_bias(inputs, stem_width, 3, strides=2, use_bias=bn_act_first, padding="same", name="stem_1_")
    nn = batchnorm_with_activation(nn, activation=activation, act_first=bn_act_first, name="stem_1_")
    nn = conv2d_no_bias(nn, stem_width, 3, strides=1, use_bias=bn_act_first, padding="same", name="stem_2_")
    # nn = batchnorm_with_activation(nn, activation=activation, name="stem_2_")

    """ stage [1, 2, 3, 4] """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    for stack_id, (num_block, out_channel, block_type) in enumerate(zip(num_blocks, out_channels, block_types)):
        is_conv_block = True if block_type[0].lower() == "c" else False
        stack_se_ratio = se_ratio[stack_id] if isinstance(se_ratio, (list, tuple)) else se_ratio
        stack_strides = strides[stack_id] if isinstance(strides, (list, tuple)) else strides
        for block_id in range(num_block):
            name = "stack_{}_block_{}_".format(stack_id + 1, block_id + 1)
            stride = stack_strides if block_id == 0 else 1
            conv_short_cut = True if block_id == 0 else False
            block_se_ratio = stack_se_ratio[block_id] if isinstance(stack_se_ratio, (list, tuple)) else stack_se_ratio
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            global_block_id += 1
            if is_conv_block:
                nn = res_MBConv(
                    nn, out_channel, conv_short_cut, stride, expansion, block_se_ratio, block_drop_rate, use_dw_strides, bn_act_first, activation, name=name
                )
            else:
                nn = res_mhsa(nn, out_channel, conv_short_cut, stride, head_dimension, block_drop_rate, activation=activation, name=name)
                nn = res_ffn(nn, expansion=expansion, drop_rate=block_drop_rate, activation=activation, name=name + "ffn_")

    nn = output_block(nn, num_classes=num_classes, drop_rate=dropout, classifier_activation=classifier_activation, act_first=bn_act_first)
    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, PRETRAINED_DICT, "coatnet", pretrained, MultiHeadRelativePositionalEmbedding)
    return model


@register_model
def CoAtNetT(input_shape=(224, 224, 3), num_classes=1000, drop_connect_rate=0, classifier_activation="softmax", **kwargs):
    num_blocks = [3, 4, 6, 3]
    out_channels = [64, 128, 256, 512]
    stem_width = 64
    return CoAtNet(**locals(), model_name="coatnett", **kwargs)


@register_model
def CoAtNet0(input_shape=(224, 224, 3), num_classes=1000, drop_connect_rate=0, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 3, 5, 2]
    out_channels = [96, 192, 384, 768]
    stem_width = 64
    return CoAtNet(**locals(), model_name="coatnet0", **kwargs)


@register_model
def CoAtNet1(input_shape=(224, 224, 3), num_classes=1000, drop_connect_rate=0.3, classifier_activation="softmax", **kwargs):
    num_blocks = [2, 6, 14, 2]
    out_channels = [96, 192, 384, 768]
    stem_width = 64
    return CoAtNet(**locals(), model_name="coatnet1", **kwargs)


@register_model
def CoAtNet2(input_shape=(224, 224, 3), num_classes=1000, drop_connect_rate=0.5, classifier_activation="softmax", **kwargs):
    num_blocks = [2, 6, 14, 2]
    out_channels = [128, 256, 512, 1024]
    stem_width = 128
    return CoAtNet(**locals(), model_name="coatnet2", **kwargs)


@register_model
def CoAtNet3(input_shape=(224, 224, 3), num_classes=1000, drop_connect_rate=0.7, classifier_activation="softmax", **kwargs):
    num_blocks = [2, 6, 14, 2]
    out_channels = [192, 384, 768, 1536]
    stem_width = 192
    return CoAtNet(**locals(), model_name="coatnet3", **kwargs)


@register_model
def CoAtNet4(input_shape=(224, 224, 3), num_classes=1000, drop_connect_rate=0.2, classifier_activation="softmax", **kwargs):
    num_blocks = [2, 12, 28, 2]
    out_channels = [192, 384, 768, 1536]
    stem_width = 192
    return CoAtNet(**locals(), model_name="coatnet4", **kwargs)


@register_model
def CoAtNet5(input_shape=(224, 224, 3), num_classes=1000, drop_connect_rate=0.2, classifier_activation="softmax", **kwargs):
    num_blocks = [2, 12, 28, 2]
    out_channels = [256, 512, 1280, 2048]
    stem_width = 192
    head_dimension = 64
    return CoAtNet(**locals(), model_name="coatnet5", **kwargs)


@register_model
def CoAtNet6(input_shape=(224, 224, 3), num_classes=1000, drop_connect_rate=0.2, classifier_activation="softmax", **kwargs):
    num_blocks = [2, 4, 8, 42, 2]
    out_channels = [192, 384, 768, 1536, 2048]
    block_types = ["conv", "conv", "conv", "transfrom", "transform"]
    strides = [2, 2, 2, 1, 2]
    stem_width = 192
    head_dimension = 128
    return CoAtNet(**locals(), model_name="coatnet6", **kwargs)


@register_model
def CoAtNet7(input_shape=(224, 224, 3), num_classes=1000, drop_connect_rate=0.2, classifier_activation="softmax", **kwargs):
    num_blocks = [2, 4, 8, 42, 2]
    out_channels = [256, 512, 1024, 2048, 3072]
    block_types = ["conv", "conv", "conv", "transfrom", "transform"]
    strides = [2, 2, 2, 1, 2]
    stem_width = 192
    head_dimension = 128
    return CoAtNet(**locals(), model_name="coatnet7", **kwargs)



import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, functional, models, initializers, image_data_format
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    conv2d_no_bias,
    group_norm,
    add_pre_post_process,
    qkv_to_multi_head_channels_last_format,
    scaled_dot_product_attention,
)
from keras_cv_attention_models.stable_diffusion.unet import res_block
from keras_cv_attention_models.download_and_load import reload_model_weights

GROUP_NORM_EPSILON = 1e-6

PRETRAINED_DICT = {"decoder": {"v1_5": "3b37d81cd07fbab875bdbb48fbb91e27"}, "encoder": {"v1_5": "d053d1d1d43cba0f357381cf2710836a"}}


def gaussian_distribution(inputs):
    # [TODO] keras.layers.GaussianNoise
    channel_axis = -1 if backend.image_data_format() == "channels_last" else 1
    mean, log_var = functional.split(inputs, 2, axis=channel_axis)
    log_var = functional.clip_by_value(log_var, -30.0, 20.0)
    std = functional.exp(log_var * 0.5)

    gaussian = initializers.RandomNormal()(std.shape)
    return mean + std * gaussian


def attention_block(inputs, num_attention_block=1, mlp_ratio=4, num_heads=4, head_dim=0, name=""):
    input_channels = inputs.shape[-1 if backend.image_data_format() == "channels_last" else 1]
    inputs_shape = functional.shape(inputs) if None in inputs.shape[1:] or -1 in inputs.shape[1:] else [-1, *inputs.shape[1:]]
    qk_scale = 1.0 / (float(input_channels) ** 0.5)

    nn = group_norm(inputs, epsilon=GROUP_NORM_EPSILON, name=name + "in_layers_")

    qq = conv2d_no_bias(nn, input_channels, use_bias=True, name=name and name + "query_")
    kk = conv2d_no_bias(nn, input_channels, use_bias=True, name=name and name + "key_")
    vv = conv2d_no_bias(nn, input_channels, use_bias=True, name=name and name + "value_")

    if image_data_format() == "channels_last":
        qq = layers.Reshape([-1, qq.shape[-1]])(qq)
        kk = functional.transpose(layers.Reshape([-1, kk.shape[-1]])(kk), [0, 2, 1])
        vv = layers.Reshape([-1, vv.shape[-1]])(vv)
    else:
        qq = functional.transpose(layers.Reshape([qq.shape[1], -1])(qq), [0, 2, 1])
        kk = layers.Reshape([kk.shape[1], -1])(kk)
        vv = layers.Reshape([vv.shape[1], -1])(vv)

    attention_scores = (qq @ kk) * qk_scale
    attention_scores = layers.Softmax(axis=-1, name=name and name + "attention_scores")(attention_scores)

    if image_data_format() == "channels_last":
        attention_output = attention_scores @ vv
        output = functional.reshape(attention_output, inputs_shape)
    else:
        attention_output = vv @ functional.transpose(attention_scores, [0, 2, 1])
        output = functional.reshape(attention_output, inputs_shape)
    output = conv2d_no_bias(output, input_channels, use_bias=True, name=name and name + "out")
    return layers.Add(name=name + "out")([output, inputs])


@register_model
def Encoder(
    input_shape=(512, 512, 3),  # input -> DownSample 3 times -> // 8, 512 -> 64
    num_blocks=[2, 2, 2, 2],
    hidden_channels=128,
    hidden_expands=[1, 2, 4, 4],
    output_channels=4,
    post_encoder_channels=4,  # > 0 value for an additional Conv2D layer after output layer
    pretrained="v1_5",
    model_name="encoder",
    kwargs=None,  # Not using, recieving parameter
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    inputs = layers.Input(backend.align_input_shape_by_image_data_format(input_shape))
    channel_axis = -1 if backend.image_data_format() == "channels_last" else 1
    nn = conv2d_no_bias(inputs, hidden_channels, kernel_size=3, use_bias=True, padding="SAME", name="stem_")

    """ Down blocks """
    for stack_id, (num_block, hidden_expand) in enumerate(zip(num_blocks, hidden_expands)):
        stack_name = "stack{}_".format(stack_id + 1)
        out_channels = hidden_expand * hidden_channels
        if stack_id > 0:
            nn = conv2d_no_bias(nn, nn.shape[channel_axis], 3, 2, use_bias=True, padding="same", use_torch_padding=False, name=stack_name + "downsample_")

        for block_id in range(num_block):
            block_name = stack_name + "down_block{}_".format(block_id + 1)
            nn = res_block(nn, time_embedding=None, out_channels=out_channels, name=block_name)
    # print(f">>>> {[ii.shape for ii in skip_connections] = }")

    """ Middle blocks """
    nn = res_block(nn, time_embedding=None, name="middle_block_1_")
    nn = attention_block(nn, name="middle_block_attn_")
    nn = res_block(nn, time_embedding=None, name="middle_block_2_")

    """ Output blocks """
    nn = group_norm(nn, epsilon=GROUP_NORM_EPSILON, name="output_")
    nn = activation_by_name(nn, activation="swish", name="output_")
    # * 2 means [mean, std]
    outputs = conv2d_no_bias(nn, output_channels * 2, kernel_size=3, use_bias=True, padding="SAME", name="output_")
    outputs = conv2d_no_bias(outputs, 2 * post_encoder_channels, use_bias=True, name="post_encoder_") if post_encoder_channels > 0 else outputs

    model = models.Model(inputs, outputs, name=model_name)
    reload_model_weights(model, PRETRAINED_DICT, "stable_diffusion", pretrained)
    return model


@register_model
def Decoder(
    input_shape=(64, 64, 4),
    num_blocks=[2, 2, 2, 2],
    hidden_channels=128,
    hidden_expands=[4, 4, 2, 1],
    output_channels=3,
    pre_decoder_channels=4,  # > 0 value for an additional Conv2D layer after input layer
    pretrained="v1_5",
    model_name="decoder",
    kwargs=None,  # Not using, recieving parameter
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    inputs = layers.Input(backend.align_input_shape_by_image_data_format(input_shape))
    channel_axis = -1 if backend.image_data_format() == "channels_last" else 1
    nn = conv2d_no_bias(inputs, pre_decoder_channels, use_bias=True, name="pre_decoder_") if pre_decoder_channels > 0 else inputs
    nn = conv2d_no_bias(nn, hidden_channels * hidden_expands[0], kernel_size=3, use_bias=True, padding="SAME", name="stem_")

    """ Middle blocks """
    nn = res_block(nn, time_embedding=None, epsilon=GROUP_NORM_EPSILON, name="middle_block_1_")
    nn = attention_block(nn, name="middle_block_attn_")
    nn = res_block(nn, time_embedding=None, epsilon=GROUP_NORM_EPSILON, name="middle_block_2_")

    """ Up blocks """
    for stack_id, (num_block, hidden_expand) in enumerate(zip(num_blocks, hidden_expands)):
        stack_name = "stack{}_".format(stack_id + 1)
        out_channels = hidden_expand * hidden_channels
        if stack_id > 0:
            nn = layers.UpSampling2D(size=2, name=stack_name + "upsample_")(nn)
            nn = conv2d_no_bias(nn, nn.shape[channel_axis], kernel_size=3, strides=1, use_bias=True, padding="same", name=stack_name + "upsample_")

        for block_id in range(num_block + 1):
            block_name = stack_name + "up_block{}_".format(block_id + 1)
            nn = res_block(nn, time_embedding=None, epsilon=GROUP_NORM_EPSILON, out_channels=out_channels, name=block_name)

    """ Output blocks """
    nn = group_norm(nn, epsilon=GROUP_NORM_EPSILON, name="output_")
    nn = activation_by_name(nn, activation="swish", name="output_")
    outputs = conv2d_no_bias(nn, output_channels, kernel_size=3, use_bias=True, padding="SAME", name="output_")

    model = models.Model(inputs, outputs, name=model_name)
    reload_model_weights(model, PRETRAINED_DICT, "stable_diffusion", pretrained)
    return model



from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    batchnorm_with_activation,
    conv2d_no_bias,
    depthwise_conv2d_no_bias,
    drop_block,
    layer_norm,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

PRETRAINED_DICT = {
    "efficientvit_b0": {"imagenet": {224: "dd0104843abccee903d93c7a5f5d3174"}},
    "efficientvit_b1": {
        "imagenet": {224: "4fac7e85f4528276fafb5bb92e3f694f", 256: "f62ff918d3805a01dd63e8c684410d19", 288: "423077ad1a76d1b3904577a3b15d0670"}
    },
    "efficientvit_b2": {
        "imagenet": {224: "b576960761ae2193ff198d8b76f0f575", 256: "9e4ff199945cfdb0a2207c5ed0d1ba96", 288: "a4cdfe86cb9dde123e70839edae123fc"}
    },
    "efficientvit_b3": {
        "imagenet": {224: "52b0266e48b7fb0f469c23f0b564ee4f", 256: "e8f3fd1ba3ce4d30d52b927d595663f9", 288: "4569c90bb93c7a657d4add91f0a7323f"}
    },
    "efficientvit_l1": {"imagenet": {224: "a487ed323d278e7b4807a4f04257ed10"}},
    "efficientvit_l2": {"imagenet": {224: "b1bc0d47aeaa90421f73a89ceac04e8b", 384: "c396388dd90e7532256a3a25fed4b5b0"}},
    "efficientvit_l3": {"imagenet": {224: "8b3d1480ce67c46101cf0a3a0fa69318", 384: "e7d5f2bce807254c24d7530553f21f6a"}},
}


def mb_conv(
    inputs, output_channel, shortcut=True, strides=1, expansion=4, is_fused=False, use_bias=False, use_norm=True, drop_rate=0, activation="hard_swish", name=""
):
    input_channel = inputs.shape[-1 if image_data_format() == "channels_last" else 1]
    if is_fused:
        nn = conv2d_no_bias(inputs, int(input_channel * expansion), 3, strides, padding="same", name=name and name + "expand_")
        nn = batchnorm_with_activation(nn, activation=activation, name=name + "expand_") if use_norm else activation_by_name(nn, activation=activation)
    elif expansion > 1:
        nn = conv2d_no_bias(inputs, int(input_channel * expansion), 1, strides=1, use_bias=use_bias, name=name + "expand_")
        nn = batchnorm_with_activation(nn, activation=activation, name=name + "expand_") if use_norm else activation_by_name(nn, activation=activation)
    else:
        nn = inputs

    if not is_fused:
        nn = depthwise_conv2d_no_bias(nn, 3, strides=strides, use_bias=use_bias, padding="same", name=name)
        nn = batchnorm_with_activation(nn, activation=activation, name=name + "dw_") if use_norm else activation_by_name(nn, activation=activation)

    pw_kernel_size = 3 if is_fused and expansion == 1 else 1
    nn = conv2d_no_bias(nn, output_channel, pw_kernel_size, strides=1, padding="same", use_bias=False, name=name + "pw_")
    nn = batchnorm_with_activation(nn, activation=None, zero_gamma=True, name=name + "pw_")
    nn = drop_block(nn, drop_rate=drop_rate, name=name)
    return layers.Add(name=name + "output")([inputs, nn]) if shortcut else layers.Activation("linear", name=name + "output")(nn)


def lite_mhsa(inputs, num_heads=8, key_dim=16, sr_ratio=5, qkv_bias=False, out_shape=None, out_bias=False, dropout=0, activation="relu", name=None):
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    input_channel = inputs.shape[channel_axis]
    height, width = inputs.shape[1:-1] if image_data_format() == "channels_last" else inputs.shape[2:]
    key_dim = key_dim if key_dim > 0 else input_channel // num_heads
    out_shape = input_channel if out_shape is None else out_shape
    emb_dim = num_heads * key_dim

    # query = layers.Dense(emb_dim, use_bias=qkv_bias, name=name and name + "query")(inputs)
    qkv = conv2d_no_bias(inputs, emb_dim * 3, use_bias=qkv_bias, name=name and name + "qkv_")
    sr_qkv = depthwise_conv2d_no_bias(qkv, kernel_size=sr_ratio, use_bias=qkv_bias, padding="same", name=name and name + "qkv_")
    sr_qkv = conv2d_no_bias(sr_qkv, emb_dim * 3, use_bias=qkv_bias, groups=3 * num_heads, name=name and name + "qkv_pw_")
    qkv = functional.concat([qkv, sr_qkv], axis=channel_axis)

    if image_data_format() == "channels_last":
        qkv = functional.reshape(qkv, [-1, height * width, qkv.shape[-1] // (3 * key_dim), 3 * key_dim])
        query, key, value = functional.split(qkv, 3, axis=-1)
        query = functional.transpose(query, [0, 2, 1, 3])
        key = functional.transpose(key, [0, 2, 3, 1])
        value = functional.transpose(value, [0, 2, 1, 3])
    else:
        qkv = functional.reshape(qkv, [-1, qkv.shape[1] // (3 * key_dim), 3 * key_dim, height * width])
        query, key, value = functional.split(qkv, 3, axis=2)
        query = functional.transpose(query, [0, 1, 3, 2])
        value = functional.transpose(value, [0, 1, 3, 2])
    query = activation_by_name(query, activation=activation)
    key = activation_by_name(key, activation=activation)
    # print(f">>>> {inputs.shape = }, {query.shape = }, {sr_ratio = }")

    query_key = query @ key
    scale = functional.reduce_sum(query_key, axis=-1, keepdims=True)
    attention_output = query_key @ value / (scale + 1e-7)  # 1e-7 for also working on float16

    if image_data_format() == "channels_last":
        output = functional.transpose(attention_output, [0, 2, 1, 3])  # [batch, q_blocks, num_heads, key_dim * attn_ratio]
        output = functional.reshape(output, [-1, height, width, output.shape[2] * output.shape[3]])
    else:
        output = functional.transpose(attention_output, [0, 1, 3, 2])  # [batch, num_heads, key_dim * attn_ratio, q_blocks]
        output = functional.reshape(output, [-1, output.shape[1] * output.shape[2], height, width])
    output = conv2d_no_bias(output, out_shape, use_bias=out_bias, name=name and name + "out_")
    output = batchnorm_with_activation(output, activation=None, name=name and name + "out_")
    return output


def EfficientViT_B(
    num_blocks=[2, 2, 3, 3],
    out_channels=[16, 32, 64, 128],
    stem_width=8,
    block_types=["conv", "conv", "transform", "transform"],
    expansions=4,  # int or list, each element in list can also be an int or list of int
    is_fused=False,  # True for L models, False for B models
    head_dimension=16,  # `num_heads = channels // head_dimension`
    output_filters=[1024, 1280],
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="hard_swish",
    drop_connect_rate=0,
    classifier_activation="softmax",
    dropout=0,
    pretrained=None,
    model_name="efficientvit",
    kwargs=None,
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)
    is_fused = is_fused if isinstance(is_fused, (list, tuple)) else ([is_fused] * len(num_blocks))

    """ stage 0, Stem_stage """
    nn = conv2d_no_bias(inputs, stem_width, 3, strides=2, padding="same", name="stem_")
    nn = batchnorm_with_activation(nn, activation=activation, name="stem_")
    nn = mb_conv(nn, stem_width, shortcut=True, expansion=1, is_fused=is_fused[0], activation=activation, name="stem_MB_")

    """ stage [1, 2, 3, 4] """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    for stack_id, (num_block, out_channel, block_type) in enumerate(zip(num_blocks, out_channels, block_types)):
        is_conv_block = True if block_type[0].lower() == "c" else False
        cur_expansions = expansions[stack_id] if isinstance(expansions, (list, tuple)) else expansions
        use_bias, use_norm = (True, False) if stack_id >= 2 else (False, True)  # fewer_norm
        cur_is_fused = is_fused[stack_id]
        for block_id in range(num_block):
            name = "stack_{}_block_{}_".format(stack_id + 1, block_id + 1)
            stride = 2 if block_id == 0 else 1
            shortcut = False if block_id == 0 else True
            cur_expansion = cur_expansions[block_id] if isinstance(cur_expansions, (list, tuple)) else cur_expansions
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks

            if is_conv_block or block_id == 0:
                cur_name = (name + "downsample_") if stride > 1 else name
                nn = mb_conv(nn, out_channel, shortcut, stride, cur_expansion, cur_is_fused, use_bias, use_norm, block_drop_rate, activation, name=cur_name)
            else:
                num_heads = out_channel // head_dimension
                attn = lite_mhsa(nn, num_heads=num_heads, key_dim=head_dimension, sr_ratio=5, name=name + "attn_")
                nn = nn + attn
                nn = mb_conv(nn, out_channel, shortcut, stride, cur_expansion, cur_is_fused, use_bias, use_norm, block_drop_rate, activation, name=name)
            global_block_id += 1

    output_filters = output_filters if isinstance(output_filters, (list, tuple)) else (output_filters, 0)
    if output_filters[0] > 0:
        nn = conv2d_no_bias(nn, output_filters[0], name="features_")
        nn = batchnorm_with_activation(nn, activation=activation, name="features_")

    if num_classes > 0:
        nn = layers.GlobalAveragePooling2D(name="avg_pool")(nn) if len(nn.shape) == 4 else nn
        if dropout > 0:
            nn = layers.Dropout(dropout, name="head_drop")(nn)
        if output_filters[1] > 0:
            nn = layers.Dense(output_filters[1], use_bias=False, name="pre_predictions")(nn)
            nn = layer_norm(nn, name="pre_predictions_")
            nn = activation_by_name(nn, activation=activation, name="pre_predictions_")
        nn = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="predictions")(nn)
    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, PRETRAINED_DICT, "efficientvit", pretrained)
    return model


@register_model
def EfficientViT_B0(input_shape=(224, 224, 3), num_classes=1000, activation="hard_swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return EfficientViT_B(**locals(), model_name="efficientvit_b0", **kwargs)


@register_model
def EfficientViT_B1(input_shape=(224, 224, 3), num_classes=1000, activation="hard_swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    out_channels = [32, 64, 128, 256]
    num_blocks = [2, 3, 4, 5]
    stem_width = 16
    output_filters = [1536, 1600]
    return EfficientViT_B(**locals(), model_name="efficientvit_b1", **kwargs)


@register_model
def EfficientViT_B2(input_shape=(224, 224, 3), num_classes=1000, activation="hard_swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    out_channels = [48, 96, 192, 384]
    num_blocks = [3, 4, 5, 7]
    stem_width = 24
    head_dimension = 32
    output_filters = [2304, 2560]
    return EfficientViT_B(**locals(), model_name="efficientvit_b2", **kwargs)


@register_model
def EfficientViT_B3(input_shape=(224, 224, 3), num_classes=1000, activation="hard_swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    out_channels = [64, 128, 256, 512]
    num_blocks = [4, 6, 7, 10]
    stem_width = 32
    head_dimension = 32
    output_filters = [2304, 2560]
    return EfficientViT_B(**locals(), model_name="efficientvit_b3", **kwargs)


""" EfficientViT_L models """


def EfficientViT_L(
    num_blocks=[2, 2, 5, 5],
    out_channels=[64, 128, 256, 512],
    stem_width=32,
    head_dimension=32,
    block_types=["conv", "conv", "conv", "transform"],
    is_fused=[True, True, False, False],
    output_filters=[3072, 3200],
    activation="gelu",
    model_name="efficientvit_l",
    **kwargs,
):
    expansions = kwargs.pop("expansions", [[16] + [4] * num_blocks[0], [16] + [4] * num_blocks[1], [16] + [4] * num_blocks[2], [24] + [6] * num_blocks[3]])
    kwargs.pop("kwargs", None)
    return EfficientViT_B(**locals(), **kwargs)


@register_model
def EfficientViT_L0(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained=None, **kwargs):
    return EfficientViT_L(**locals(), model_name="efficientvit_l0", **kwargs)


@register_model
def EfficientViT_L1(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 2, 7, 7]
    return EfficientViT_L(**locals(), model_name="efficientvit_l1", **kwargs)


@register_model
def EfficientViT_L2(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 3, 9, 9]
    return EfficientViT_L(**locals(), model_name="efficientvit_l2", **kwargs)


@register_model
def EfficientViT_L3(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 3, 9, 9]
    out_channels = [128, 256, 512, 1024]
    stem_width = 64
    output_filters = [6144, 6400]
    return EfficientViT_L(**locals(), model_name="efficientvit_l3", **kwargs)



from keras_cv_attention_models.fastvit.fastvit import (
    FastViT,
    FastViT_T8,
    FastViT_T12,
    FastViT_S12,
    FastViT_SA12,
    FastViT_SA24,
    FastViT_SA36,
    FastViT_MA36,
    # switch_to_deploy,
)

__head_doc__ = """
Keras implementation of [Github NVlabs/FasterViT](https://github.com/NVlabs/FasterViT).
Paper [PDF 2306.06189 FasterViT: Fast Vision Transformers with Hierarchical Attention](https://arxiv.org/pdf/2306.06189.pdf).
"""

__tail_doc__ = """  block_types: block types for each stack,
      - `conv` or any `c` / `C` starts word, means `rep_conv_block` block.
      - `transfrom` or any not `c` / `C` starts word, means `multi_head_self_attention` block.
      value could be in format like `"cctt"` or `"CCTT"` or `["conv", "conv", "transfrom", "transform"]`.
      `["conv", "conv", "conv", "conv"]` for SA models, all conv for others.
  layer_scale: layer scale init value, [Going deeper with Image Transformers](https://arxiv.org/abs/2103.17239).
      Default 1e-6 for SA models, 1e-5 for others.
  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  deploy: boolean value if build a fused model. **Evaluation only, not good for training**.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `hard_swish`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be a constant value like `0.2`,
      or a tuple value like `(0, 0.2)` indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  dropout: top dropout rate if top layers is included. Default 0.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `None`.
  pretrained: one of `None` (random initialization) or 'imagenet' (pre-training on ImageNet).
      Will try to download and load pre-trained model weights if not None.
  **kwargs: other parameters if available.

Returns:
    A `keras.Model` instance.
"""

FastViT.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of block for each stack.
  out_channels: output channels for each stack.
  stem_width: channel dimension output for stem block, default -1 for using out_channels[0].
  mlp_ratio: int value for mlp_ratio for each stack.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model        | Params | FLOPs | Input | Top1 Acc |
  | ------------ | ------ | ----- | ----- | -------- |
  | FastViT_T8   | 4.03M  | 0.65G | 256   | 76.2     |
  | - distill    | 4.03M  | 0.65G | 256   | 77.2     |
  | FastViT_T12  | 7.55M  | 1.34G | 256   | 79.3     |
  | - distill    | 7.55M  | 1.34G | 256   | 80.3     |
  | FastViT_S12  | 9.47M  | 1.74G | 256   | 79.9     |
  | - distill    | 9.47M  | 1.74G | 256   | 81.1     |
  | FastViT_SA12 | 11.58M | 1.88G | 256   | 80.9     |
  | - distill    | 11.58M | 1.88G | 256   | 81.9     |
  | FastViT_SA24 | 21.55M | 3.66G | 256   | 82.7     |
  | - distill    | 21.55M | 3.66G | 256   | 83.4     |
  | FastViT_SA36 | 31.53M | 5.44G | 256   | 83.6     |
  | - distill    | 31.53M | 5.44G | 256   | 84.2     |
  | FastViT_MA36 | 44.07M | 7.64G | 256   | 83.9     |
  | - distill    | 44.07M | 7.64G | 256   | 84.6     |
"""

FastViT_T8.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

FastViT_T12.__doc__ = FastViT_T8.__doc__
FastViT_S12.__doc__ = FastViT_T8.__doc__
FastViT_SA12.__doc__ = FastViT_T8.__doc__
FastViT_SA24.__doc__ = FastViT_T8.__doc__
FastViT_SA36.__doc__ = FastViT_T8.__doc__
FastViT_MA36.__doc__ = FastViT_T8.__doc__



import math
import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, functional, models, image_data_format
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    conv2d_no_bias,
    depthwise_conv2d_no_bias,
    batchnorm_with_activation,
    drop_block,
    make_divisible,
    se_module,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

BATCH_NORM_EPSILON = 1e-5

PRETRAINED_DICT = {
    "repvit_m1": {"imagenet": "9f28080702da8feac6845a390a4eacfe"},
    "repvit_m2": {"imagenet": "db556c2d3c91808b77417e3a878450a0"},
    "repvit_m3": {"imagenet": "39d803c1c03da4f14fe6c09086663725"},
}


def rep_vgg_depthwise(inputs, kernel_size=3, strides=1, deploy=False, name=""):
    if deploy:
        return depthwise_conv2d_no_bias(inputs, kernel_size=kernel_size, strides=strides, use_bias=True, padding="same", name=name + "REPARAM_1_")

    dw_1 = depthwise_conv2d_no_bias(inputs, kernel_size=kernel_size, strides=strides, padding="same", name=name + "REPARAM_1_")
    dw_1 = batchnorm_with_activation(dw_1, epsilon=BATCH_NORM_EPSILON, activation=None, name=name + "REPARAM_1_")
    dw_2 = depthwise_conv2d_no_bias(inputs, 1, name=name + "REPARAM_2_")
    dw_2 = batchnorm_with_activation(dw_2, epsilon=BATCH_NORM_EPSILON, activation=None, name=name + "REPARAM_2_")
    return layers.Add(name=name + "REPARAM_out")([dw_1, dw_2, inputs])


def dwconv_bn_conv_bn(inputs, out_channel, kernel_size=1, strides=1, deploy=False, name=""):
    nn = depthwise_conv2d_no_bias(inputs, kernel_size, strides=strides, use_bias=deploy, padding="same", name=name + "1_")
    nn = nn if deploy else batchnorm_with_activation(nn, epsilon=BATCH_NORM_EPSILON, name=name + "1_")
    nn = conv2d_no_bias(nn, out_channel, use_bias=deploy, name=name + "2_")
    nn = nn if deploy else batchnorm_with_activation(nn, epsilon=BATCH_NORM_EPSILON, activation=None, name=name + "2_")
    return nn


def conv_bn_act_conv_bn(
    inputs, hidden_channel, out_channel, use_residual=True, kernel_size=1, strides=1, deploy=False, drop_rate=0, activation="gelu", name=""
):
    nn = conv2d_no_bias(inputs, hidden_channel, kernel_size, strides=strides, use_bias=deploy, padding="same", name=name + "1_")
    nn = nn if deploy else batchnorm_with_activation(nn, epsilon=BATCH_NORM_EPSILON, activation=None, name=name + "1_")
    nn = activation_by_name(nn, activation=activation, name=name + "1_")
    nn = conv2d_no_bias(nn, out_channel, kernel_size, strides=strides, use_bias=deploy, padding="same", name=name + "2_")
    nn = nn if deploy else batchnorm_with_activation(nn, epsilon=BATCH_NORM_EPSILON, activation=None, name=name + "2_")
    nn = drop_block(nn, drop_rate=drop_rate, name=name)
    return (nn + inputs) if use_residual else nn


def RepViT(
    num_blocks=[3, 3, 15, 2],
    out_channels=[48, 96, 192, 384],
    stem_width=-1,
    se_ratio=0.25,  # will use `se_module` every other block in each stack if > 0
    input_shape=(224, 224, 3),
    deploy=False,  # build model with rep_vgg_depthwise/conv+bn/distill_head all being fused
    num_classes=1000,
    activation="gelu",
    drop_connect_rate=0,
    classifier_activation="softmax",
    use_distillation=False,
    dropout=0,
    pretrained=None,
    model_name="repvit",
    kwargs=None,
):
    """Patch stem"""
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)
    stem_width = stem_width if stem_width > 0 else out_channels[0]
    nn = conv_bn_act_conv_bn(
        inputs, stem_width // 2, stem_width, use_residual=False, kernel_size=3, strides=2, deploy=deploy, activation=activation, name="stem_"
    )

    """ stage [1, 2, 3, 4] """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    for stack_id, (num_block, out_channel) in enumerate(zip(num_blocks, out_channels)):
        out_channel = make_divisible(out_channel, 8)
        stack_name = "stack{}_".format(stack_id + 1)
        if stack_id > 0:
            nn = dwconv_bn_conv_bn(nn, out_channel, kernel_size=3, strides=2, deploy=deploy, name=stack_name + "downsample_token_mixer_")
            nn = conv_bn_act_conv_bn(nn, out_channel * 2, out_channel, deploy=deploy, activation=activation, name=stack_name + "downsample_channel_mixer_")

        for block_id in range(num_block):
            name = "stack{}_block{}_".format(stack_id + 1, block_id + 1)
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            cur_se_ratio = 0 if block_id % 2 == 1 or block_id == (num_block - 1) else se_ratio
            # print(f"{name = }, {cur_se_ratio = }")
            nn = rep_vgg_depthwise(nn, deploy=deploy, name=name + "repvgg_")
            nn = se_module(nn, cur_se_ratio, activation="relu", name=name + "se_") if cur_se_ratio > 0 else nn
            nn = conv_bn_act_conv_bn(
                nn, out_channel * 2, out_channel, deploy=deploy, activation=activation, drop_rate=block_drop_rate, name=name + "channel_mixer_"
            )
            global_block_id += 1

    if deploy and num_classes > 0:
        nn = layers.GlobalAveragePooling2D()(nn)
        out = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="head")(nn)
    elif num_classes > 0:
        nn = layers.GlobalAveragePooling2D(keepdims=True)(nn)
        out = batchnorm_with_activation(nn, activation=None, name="head_")
        out = layers.Flatten()(out)
        out = layers.Dropout(dropout, name="head_drop")(out) if dropout > 0 else out
        out = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="head")(out)

        if use_distillation:
            distill = batchnorm_with_activation(nn, activation=None, name="distill_head_")
            distill = layers.Flatten()(distill)
            distill = layers.Dropout(dropout, name="distill_head_drop")(distill) if dropout > 0 else distill
            distill = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="distill_head")(distill)
            out = [out, distill]
    else:
        out = nn
    model = models.Model(inputs, out, name=model_name)
    reload_model_weights(model, PRETRAINED_DICT, "repvit", pretrained)

    add_pre_post_process(model, rescale_mode="torch")
    model.switch_to_deploy = lambda: switch_to_deploy(model)
    return model


def switch_to_deploy(model):
    from keras_cv_attention_models.model_surgery.model_surgery import fuse_reparam_blocks, convert_to_fused_conv_bn_model, fuse_distill_head

    new_model = fuse_distill_head(model) if "head" in model.output_names else model
    new_model = fuse_reparam_blocks(convert_to_fused_conv_bn_model(new_model))
    add_pre_post_process(new_model, rescale_mode=model.preprocess_input.rescale_mode, post_process=model.decode_predictions)
    return new_model


@register_model
def RepViT_M1(
    input_shape=(224, 224, 3), num_classes=1000, deploy=False, use_distillation=False, classifier_activation="softmax", pretrained="imagenet", **kwargs
):
    return RepViT(**locals(), model_name="repvit_m1" + ("_deploy" if deploy else ""), **kwargs)


@register_model
def RepViT_M2(
    input_shape=(224, 224, 3), num_classes=1000, deploy=False, use_distillation=False, classifier_activation="softmax", pretrained="imagenet", **kwargs
):
    num_blocks = [3, 3, 13, 2]
    out_channels = [64, 128, 256, 512]
    return RepViT(**locals(), model_name="repvit_m2" + ("_deploy" if deploy else ""), **kwargs)


@register_model
def RepViT_M3(
    input_shape=(224, 224, 3), num_classes=1000, deploy=False, use_distillation=False, classifier_activation="softmax", pretrained="imagenet", **kwargs
):
    num_blocks = [5, 5, 19, 2]
    out_channels = [64, 128, 256, 512]
    return RepViT(**locals(), model_name="repvit_m3" + ("_deploy" if deploy else ""), **kwargs)



from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.download_and_load import reload_model_weights
from keras_cv_attention_models.attention_layers import batchnorm_with_activation, conv2d_no_bias, drop_block, quad_stem, add_pre_post_process


PRETRAINED_DICT = {
    "resnet51q": {"imagenet": "2b54c5e252bd58f37454e6fb273716f7"},
}


def quad_block(inputs, filters, groups_div=32, strides=1, conv_shortcut=False, expansion=4, extra_conv=False, drop_rate=0, activation="swish", name=""):
    expanded_filter = filters * expansion
    groups = filters // groups_div if groups_div != 0 else 1
    if conv_shortcut:
        shortcut = conv2d_no_bias(inputs, expanded_filter, 1, strides=strides, name=name + "shortcut_")
        shortcut = batchnorm_with_activation(shortcut, activation=None, zero_gamma=False, name=name + "shortcut_")
    else:
        shortcut = inputs

    if groups != 1:  # Edge block
        nn = conv2d_no_bias(inputs, filters, 1, strides=1, padding="valid", name=name + "1_")
        nn = batchnorm_with_activation(nn, activation=activation, zero_gamma=False, name=name + "1_")
    else:
        nn = inputs

    nn = conv2d_no_bias(nn, filters, 3, strides=strides, padding="same", groups=groups, name=name + "groups_")
    nn = batchnorm_with_activation(nn, activation=activation, zero_gamma=False, name=name + "2_")

    if extra_conv:
        nn = conv2d_no_bias(nn, filters, 3, strides=1, padding="same", groups=groups, name=name + "extra_groups_")
        nn = batchnorm_with_activation(nn, activation=activation, zero_gamma=False, name=name + "extra_2_")

    nn = conv2d_no_bias(nn, expanded_filter, 1, strides=1, padding="valid", name=name + "3_")
    nn = batchnorm_with_activation(nn, activation=None, zero_gamma=True, name=name + "3_")

    # print(">>>> shortcut:", shortcut.shape, "nn:", nn.shape)
    nn = drop_block(nn, drop_rate)
    nn = layers.Add(name=name + "add")([shortcut, nn])
    return layers.Activation(activation, name=name + "output")(nn)


def quad_stack(inputs, blocks, filters, groups_div, strides=2, expansion=4, extra_conv=False, stack_drop=0, activation="swish", name=""):
    nn = inputs
    stack_drop_s, stack_drop_e = stack_drop if isinstance(stack_drop, (list, tuple)) else [stack_drop, stack_drop]
    for id in range(blocks):
        conv_shortcut = True if id == 0 and (strides != 1 or inputs.shape[-1] != filters * expansion) else False
        cur_strides = strides if id == 0 else 1
        block_name = name + "block{}_".format(id + 1)
        block_drop_rate = stack_drop_s + (stack_drop_e - stack_drop_s) * id / blocks
        nn = quad_block(nn, filters, groups_div, cur_strides, conv_shortcut, expansion, extra_conv, block_drop_rate, activation, name=block_name)
    return nn


def ResNetQ(
    num_blocks,
    out_channels=[64, 128, 384, 384],
    stem_width=128,
    stem_act=False,
    expansion=4,
    groups_div=32,
    extra_conv=False,
    num_features=2048,
    strides=2,
    stem_downsample=False,
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="swish",
    drop_connect_rate=0,
    classifier_activation="softmax",
    pretrained="imagenet",
    model_name="resnetq",
    kwargs=None,
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(shape=input_shape)
    nn = quad_stem(inputs, stem_width, activation=activation, stem_act=stem_act, name="stem_")
    nn = batchnorm_with_activation(nn, activation=activation, name="stem_")
    if stem_downsample:
        nn = layers.ZeroPadding2D(padding=1, name="stem_pool_pad")(nn)
        nn = layers.MaxPooling2D(pool_size=3, strides=2, name="stem_pool")(nn)

    total_blocks = sum(num_blocks)
    global_block_id = 0
    drop_connect_s, drop_connect_e = 0, drop_connect_rate
    strides = strides if isinstance(strides, (list, tuple)) else [1, 2, 2, strides]
    for id, (num_block, out_channel, stride) in enumerate(zip(num_blocks, out_channels, strides)):
        name = "stack{}_".format(id + 1)
        stack_drop_s = drop_connect_rate * global_block_id / total_blocks
        stack_drop_e = drop_connect_rate * (global_block_id + num_block) / total_blocks
        stack_drop = (stack_drop_s, stack_drop_e)
        cur_expansion = expansion[id] if isinstance(expansion, (list, tuple)) else expansion
        cur_extra_conv = extra_conv[id] if isinstance(extra_conv, (list, tuple)) else extra_conv
        cur_groups_div = groups_div[id] if isinstance(groups_div, (list, tuple)) else groups_div
        nn = quad_stack(nn, num_block, out_channel, cur_groups_div, stride, cur_expansion, cur_extra_conv, stack_drop, activation, name=name)
        global_block_id += num_block

    if num_features != 0:  # efficientnet like
        nn = conv2d_no_bias(nn, num_features, 1, strides=1, name="features_")
        nn = batchnorm_with_activation(nn, activation=activation, name="features_")

    if num_classes > 0:
        nn = layers.GlobalAveragePooling2D(name="avg_pool")(nn)
        nn = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="predictions")(nn)

    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, pretrained_dict=PRETRAINED_DICT, sub_release="resnet_family", pretrained=pretrained)
    return model


@register_model
def ResNet51Q(input_shape=(224, 224, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 4, 6, 4]
    out_channels = [64, 128, 384, 384 * 4]
    stem_width = 128
    stem_act = False
    expansion = [4, 4, 4, 1]
    groups_div = [32, 32, 32, 1]
    extra_conv = False
    num_features = 2048
    return ResNetQ(**locals(), model_name="resnet51q", **kwargs)


@register_model
def ResNet61Q(input_shape=(224, 224, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [1, 4, 6, 4]
    out_channels = [256, 128, 384, 384 * 4]
    stem_width = 128
    stem_act = True
    expansion = [1, 4, 4, 1]
    groups_div = [0, 32, 32, 1]
    extra_conv = [False, True, True, True]
    num_features = 2048
    return ResNetQ(**locals(), model_name="resnet61q", **kwargs)



import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.aotnet import AotNet
from keras_cv_attention_models.download_and_load import reload_model_weights
from keras_cv_attention_models.attention_layers import batchnorm_with_activation, conv2d_no_bias

PRETRAINED_DICT = {
    "resnest101": {"imagenet": "63f9ebdcd32529cbc4b4fbbec3d1bb2f"},
    "resnest200": {"imagenet": "8e211dcb089b588e18d36ba7cdf92ef0"},
    "resnest269": {"imagenet": "4309ed1b0a8ae92f2b1143dc3512c5c7"},
    "resnest50": {"imagenet": "eee7b20a229821f730ab205b6afeb369"},
}


def rsoftmax(inputs, groups):
    if groups > 1:
        nn = functional.reshape(inputs, [-1, groups, int(np.prod(inputs.shape[1:])) // groups])
        # nn = tf.transpose(nn, [0, 2, 1, 3])
        nn = functional.softmax(nn, axis=1)
        nn = functional.reshape(nn, [-1, *inputs.shape[1:]])
    else:
        nn = layers.Activation("sigmoid")(inputs)
    return nn


def split_attention_conv2d(inputs, filters, kernel_size=3, strides=1, downsample_first=False, groups=2, activation="relu", name=""):
    h_axis, w_axis = [1, 2] if image_data_format() == "channels_last" else [2, 3]
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    in_channels = inputs.shape[channel_axis]
    conv_strides = strides if downsample_first else 1
    if groups == 1:
        logits = conv2d_no_bias(inputs, filters, kernel_size, strides=conv_strides, padding="same", name=name and name + "1_")
    else:
        # Using groups=2 is slow in `mixed_float16` policy
        # logits = conv2d_no_bias(inputs, filters * groups, kernel_size, padding="same", groups=groups, name=name and name + "1_")
        logits = []
        splitted_inputs = functional.split(inputs, groups, axis=channel_axis)
        for ii in range(groups):
            conv_name = name and name + "1_g{}_".format(ii + 1)
            logits.append(conv2d_no_bias(splitted_inputs[ii], filters, kernel_size, strides=conv_strides, padding="same", name=conv_name))
        logits = functional.concat(logits, axis=channel_axis)
    logits = batchnorm_with_activation(logits, activation=activation, name=name and name + "1_")

    if groups > 1:
        splited = functional.split(logits, groups, axis=channel_axis)
        gap = functional.reduce_sum(splited, axis=0)
    else:
        gap = logits
    gap = functional.reduce_mean(gap, [h_axis, w_axis], keepdims=True)

    reduction_factor = 4
    inter_channels = max(in_channels * groups // reduction_factor, 32)
    atten = layers.Conv2D(inter_channels, kernel_size=1, name=name and name + "2_conv")(gap)
    atten = batchnorm_with_activation(atten, activation=activation, name=name and name + "2_")
    atten = layers.Conv2D(filters * groups, kernel_size=1, name=name and name + "3_conv")(atten)
    atten = rsoftmax(atten, groups)
    out = layers.Multiply()([atten, logits])

    if groups > 1:
        out = functional.split(out, groups, axis=channel_axis)
        out = functional.reduce_sum(out, axis=0)

    if not downsample_first and strides > 1:
        out = layers.ZeroPadding2D(padding=1, name=name and name + "pool_pad")(out)
        out = layers.AvgPool2D(3, strides=2, name=name and name + "pool")(out)
    return out


def ResNest(
    num_blocks=[3, 4, 6, 3],
    stem_width=64,
    stem_type="deep",
    attn_types="sa",
    bn_after_attn=False,
    shortcut_type="avg",
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="relu",
    classifier_activation="softmax",
    pretrained="imagenet",
    model_name="",
    **kwargs
):
    kwargs.pop("kwargs", None)
    model = AotNet(**locals(), **kwargs)
    reload_model_weights(model, pretrained_dict=PRETRAINED_DICT, sub_release="resnest", pretrained=pretrained)
    return model


@register_model
def ResNest50(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", groups=2, **kwargs):
    return ResNest(num_blocks=[3, 4, 6, 3], stem_width=64, model_name="resnest50", **locals(), **kwargs)


@register_model
def ResNest101(input_shape=(256, 256, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", groups=2, **kwargs):
    return ResNest(num_blocks=[3, 4, 23, 3], stem_width=128, model_name="resnest101", **locals(), **kwargs)


@register_model
def ResNest200(input_shape=(320, 320, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", groups=2, **kwargs):
    return ResNest(num_blocks=[3, 24, 36, 3], stem_width=128, model_name="resnest200", **locals(), **kwargs)


@register_model
def ResNest269(input_shape=(416, 416, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", groups=2, **kwargs):
    return ResNest(num_blocks=[3, 30, 48, 8], stem_width=128, model_name="resnest269", **locals(), **kwargs)



import numpy as np
import tensorflow as tf
from keras_cv_attention_models.plot_func import get_plot_cols_rows, stack_and_plot_images


def matmul_prod(aa):
    vv = aa[0]
    for ii in aa[1:]:
        vv = np.matmul(vv, ii)
    return vv


def apply_mask_2_image(image, mask):
    if len(mask.shape) == 1:
        width = height = int(np.sqrt(mask.shape[0]))
        mask = mask[-width * height :]
    else:
        height, width = mask.shape[:2]
    mask = mask.reshape(width, height, 1)
    mask = tf.image.resize(mask / mask.max(), image.shape[:2], method="bilinear").numpy()
    return (mask * image).astype("uint8")


def clip_max_value_matrix(dd, axis=0):
    # print("Before:", dd.max())
    for ii in range(dd.shape[axis]):
        if axis == 0:
            max_idx = np.argmax(dd[ii])
            dd[ii, max_idx] = dd[ii].min()
            dd[ii, max_idx] = dd[ii].max()
        else:
            max_idx = np.argmax(dd[:, ii])
            dd[max_idx, ii] = dd[:, ii].min()
            dd[max_idx, ii] = dd[:, ii].max()
    # print("After:", dd.max())
    return dd


def down_sample_matrix_axis_0(dd, target, method="avg"):
    if dd.shape[0] == target:
        return dd
    rate = int(np.sqrt(dd.shape[0] // target))
    hh = ww = int(np.sqrt(dd.shape[0]))
    dd = dd[: hh * ww].reshape(1, hh, ww, -1)
    if rate == 0:  # Upsample
        hh = ww = int(np.sqrt(target))
        dd = tf.image.resize(dd, [hh, ww]).numpy()
    elif "avg" in method.lower():
        dd = tf.nn.avg_pool(dd, rate, rate, "valid").numpy()
    elif "swin" in method.lower():
        dd = dd.reshape(1, hh // 2, 2, ww // 2, 2, -1).transpose(0, 1, 3, 4, 2, 5)
    else:
        dd = tf.nn.max_pool(dd, rate, rate, "valid").numpy()
    dd = dd.reshape(-1, dd.shape[-1])
    return dd


def plot_attention_score_maps(model, image, rescale_mode="auto", attn_type="auto", rows=-1, base_size=3):
    import matplotlib.pyplot as plt
    from keras_cv_attention_models.common_layers import PreprocessInput

    if rescale_mode.lower() == "auto":
        rescale_mode = getattr(model, "rescale_mode", "torch")
        print(">>>> rescale_mode:", rescale_mode)

    if isinstance(model, tf.keras.models.Model):
        imm_inputs = PreprocessInput(input_shape=model.input_shape[1:3], rescale_mode=rescale_mode)(image)
        try:
            pred = model(imm_inputs).numpy()
            if model.layers[-1].activation.__name__ != "softmax":
                pred = tf.nn.softmax(pred).numpy()  # If classifier activation is not softmax
            print(">>>> Prediction:", tf.keras.applications.imagenet_utils.decode_predictions(pred)[0])
        except:
            pass
        outputs = [ii.output for ii in model.layers if ii.name.endswith("attention_scores") and "channel_attn_attention_scores" not in ii.name]
        bb = tf.keras.models.Model(model.inputs[0], outputs)
        attn_scores = bb(imm_inputs)
        layer_name_title = "\nLayer name: {} --> {}".format(bb.output_names[-1], bb.output_names[0])
    else:
        attn_scores = model
        layer_name_title = ""
        assert attn_type != "auto"
    print(">>>> attn_scores:", {name: out.shape.as_list() for name, out in zip(bb.output_names, attn_scores)})

    attn_type = attn_type.lower()
    check_type_is = lambda tt: (tt in model.name.lower()) if attn_type == "auto" else (attn_type.startswith(tt))
    if check_type_is("beit"):
        # beit attn_score [batch, num_heads, cls_token + hh * ww, cls_token + hh * ww]
        print(">>>> Attention type: beit")
        mask = [np.array(ii)[0].mean(0) + np.eye(ii.shape[-1]) for ii in attn_scores][::-1]
        mask = [(ii / ii.sum()) for ii in mask]
        cum_mask = [matmul_prod(mask[: ii + 1])[0] for ii in range(len(mask))]
        mask = [ii[0] for ii in mask]
    elif check_type_is("mobilevit_v2"):
        # mobilevit_v2 attn_score [batch, patch_size * patch_size, patch_hh * patch_ww, 1]
        print(">>>> Attention type: mobilevit_v2")
        mask = [np.array(ii)[0].mean((0)) for ii in attn_scores if len(ii.shape) == 4][::-1]
        cum_mask = [down_sample_matrix_axis_0(mask[ii], mask[-1].shape[0]) for ii in range(len(mask) - 1)] + [mask[-1]]
        cum_mask = [np.prod(np.stack(cum_mask[: ii + 1]), axis=0).mean(-1) for ii in range(len(cum_mask))]
        mask = [ii.mean(-1) for ii in mask]
    elif check_type_is("bot") or check_type_is("cmt") or check_type_is("mobilevit"):
        # bot attn_score [batch, num_heads, hh * ww, hh * ww]
        print(">>>> Attention type: bot / cmt / mobilevit")
        mask = [np.array(ii)[0].mean((0)) for ii in attn_scores if len(ii.shape) == 4][::-1]
        mask = [clip_max_value_matrix(ii) for ii in mask]  # Or it will be too dark.
        cum_mask = [mask[0]] + [down_sample_matrix_axis_0(mask[ii], mask[ii - 1].shape[1], "avg") for ii in range(1, len(mask))]
        cum_mask = [matmul_prod(cum_mask[: ii + 1]).mean(0) for ii in range(len(cum_mask))]
        mask = [ii.mean(0) for ii in mask]
    elif check_type_is("coatnet") or check_type_is("uniformer") or check_type_is("swin") or check_type_is("davit"):
        # coatnet attn_score [batch, num_heads, hh * ww, hh * ww]
        print(">>>> Attention type: coatnet / uniformer / swin / davit")
        mask = [np.array(ii)[0].mean((0)) for ii in attn_scores if len(ii.shape) == 4][::-1]
        downsample_method = "swin" if check_type_is("swin") else "max"
        cum_mask = [mask[0]] + [down_sample_matrix_axis_0(mask[ii], mask[ii - 1].shape[1], downsample_method) for ii in range(1, len(mask))]
        cum_mask = [matmul_prod(cum_mask[: ii + 1]).mean(0) for ii in range(len(cum_mask))]
        mask = [ii.mean(0) for ii in mask]
    elif check_type_is("coat"):
        # coat attn_score [batch, num_heads, cls_token + hh * ww, key_dim]
        print(">>>> Attention type: coat")
        mask = [np.array(ii)[0].mean((0))[1:] for ii in attn_scores if len(ii.shape) == 4][::-1]
        mask = [ii.max(-1, keepdims=True) for ii in mask]
        target_shape = np.min([ii.shape[0] for ii in mask])
        cum_mask = [down_sample_matrix_axis_0(mask[ii], target_shape, "max") for ii in range(len(mask))]
        cum_mask = [ii[:, 0] for ii in cum_mask]
        mask = [ii[:, 0] for ii in mask]
    elif check_type_is("levit"):
        # levit attn_score [batch, num_heads, q_blocks, k_blocks]
        print(">>>> Attention type: levit")
        mask = [np.array(ii)[0].mean(0) for ii in attn_scores][::-1]
        cum_mask = [matmul_prod(mask[: ii + 1]).mean(0) for ii in range(len(mask))]
        mask = [ii.mean(0) for ii in mask]
    elif check_type_is("halo"):
        # halo attn_score [batch, num_heads, hh, ww, query_block * query_block, kv_kernel * kv_kernel]
        print(">>>> Attention type: halo")
        from einops import rearrange
        from keras_cv_attention_models.attention_layers import CompatibleExtractPatches

        mask = [np.array(ii)[0].mean(0) for ii in attn_scores if len(ii.shape) == 6][::-1]

        qqs = [int(np.sqrt(ii.shape[2])) for ii in mask]  # query_kernel
        vvs = [int(np.sqrt(ii.shape[3])) for ii in mask]  # kv_kernel
        hhs = [(jj - ii) // 2 for ii, jj in zip(qqs, vvs)]  # halo_size
        tt = [rearrange(ii, "hh ww (hb wb) cc -> (hh hb) (ww wb) cc", hb=qq, wb=qq) for ii, qq in zip(mask, qqs)]
        tt = [tf.expand_dims(tf.pad(ii, [[hh, hh], [hh, hh], [0, 0]]), 0) for ii, hh in zip(tt, hhs)]
        tt = [CompatibleExtractPatches(vv, qq, padding="valid", compressed=False)(ii).numpy()[0] for ii, vv, qq in zip(tt, vvs, qqs)]
        tt = [rearrange(ii, "hh ww hb wb cc -> hh ww (hb wb) cc").mean((0, 1)) for ii in tt]
        # tt = [tf.reduce_max(rearrange(ii, "hh ww hb wb cc -> hh ww (hb wb) cc"), axis=(0, 1)).numpy() for ii in tt]
        cum_mask = [matmul_prod(tt[: ii + 1]).mean(0) for ii in range(len(tt))]
        mask = [ii.mean((0, 1, 2)) for ii in mask]
    else:
        print(">>>> Attention type: cot / volo / unknown")
        # cot attn_score [batch, 1, 1, filters, randix]
        # volo attn_score [batch, hh, ww, num_heads, kernel_size * kernel_size, kernel_size * kernel_size]
        print("[{}] still don't know how...".format(attn_type))
        return None, None, None

    masked_image = [apply_mask_2_image(image, ii) for ii in mask]
    cum_masked_image = [apply_mask_2_image(image, ii) for ii in cum_mask]

    cols, rows = get_plot_cols_rows(len(mask), rows)
    fig, axes = plt.subplots(2, 1, figsize=(base_size * cols, base_size * rows * 2))
    stack_and_plot_images(masked_image, margin=5, rows=rows, ax=axes[0])
    axes[0].set_title("Attention scores: attn_scores[{}] --> attn_scores[0]".format(len(mask)) + layer_name_title)
    stack_and_plot_images(cum_masked_image, margin=5, rows=rows, ax=axes[1])
    axes[1].set_title("Accumulated attention scores: attn_scores[{}:] --> attn_scores[0:]".format(len(mask) - 1) + layer_name_title)
    fig.tight_layout()
    return mask, cum_mask, fig



import math
from keras_cv_attention_models.backend import layers, models, functional, initializers
from keras_cv_attention_models import attention_layers


def text_model_index_header(text_input, text_output, latents_dim=512, eot_token=None):
    if eot_token is None:
        eol_index = functional.argmax(functional.pad(text_input[:, 1:], [[0, 0], [1, 0]]), axis=-1)  # Skip <|startoftext|> no matter what it is
    else:
        assert eot_token != 0, "eot_token is 0, conflict with the padding value"
        eol_index = functional.argmax(functional.equal(functional.pad(text_input[:, 1:], [[0, 0], [1, 0]]), eot_token), axis=-1)

    text_output = functional.gather_nd(text_output, functional.expand_dims(eol_index, axis=-1), batch_dims=1)
    if latents_dim > 0:
        kernel_initializer = initializers.RandomNormal(stddev=text_output.shape[-1] ** -0.5)
        text_output = layers.Dense(latents_dim, use_bias=False, kernel_initializer=kernel_initializer, dtype="float32", name="text_latents")(text_output)
    return text_output


def add_text_model_index_header(model, latents_dim=512, caption_tokenizer=None):
    # Use None for argmax as eot position for GPT2Tokenizer and SimpleTokenizer. For SentencePieceTokenizer eot==2, pass the actual value
    eot_token = None if caption_tokenizer is None or caption_tokenizer.eot_token >= caption_tokenizer.vocab_size - 2 else caption_tokenizer.eot_token
    text_input, text_output = model.inputs[0], model.outputs[-1]
    text_output = text_model_index_header(text_input, text_output, latents_dim=latents_dim, eot_token=eot_token)
    return models.Model(text_input, text_output, name=model.name)


def convert_to_clip_model(image_model, text_model=None, caption_tokenizer=None, epsilon=1e-6):
    """
    >>> import tensorflow as tf  # Or import torch for PyTroch backend
    >>> from keras_cv_attention_models import clip, gpt2, beit
    >>> image_model = beit.ViT(num_classes=512, classifier_activation=None)
    >>> text_model = gpt2.GPT2_Base(include_top=False)
    >>> caption_tokenizer = clip.SimpleTokenizer()
    >>> model, image_model, text_model = clip.convert_to_clip_model(image_model, text_model, caption_tokenizer)
    >>> model.run_prediction(tf.ones([1, 224, 224, 3]), ['cat', 'dog'])
    """
    if text_model is None:
        print(">>>> Build text_model from image_model")
        kwargs = {"vocab_size": caption_tokenizer.vocab_size} if caption_tokenizer is not None and hasattr(caption_tokenizer, "vocab_size") else {}
        image_model, text_model = build_text_model_from_image_model(image_model, **kwargs)

    image_input, image_output = image_model.inputs[0], image_model.outputs[-1]
    if text_model.output_names[0] != "text_latents":
        text_model = add_text_model_index_header(text_model, latents_dim=image_output.shape[-1], caption_tokenizer=caption_tokenizer)
    text_input, text_output = text_model.inputs[0], text_model.outputs[-1]

    # Return similarity directly for avoiding re-calculating in metrics again
    # text_latents = text_output / functional.norm(text_output, ord="euclidean", axis=-1, keepdims=True)
    text_latents = text_output * functional.rsqrt(functional.reduce_sum(functional.square(text_output), axis=-1, keepdims=True) + epsilon)
    image_latents = layers.Identity(dtype="float32", name="image_latents")(image_output)  # Give a name for locating this layer in split
    # image_latents = image_latents / functional.norm(image_latents, ord="euclidean", axis=-1, keepdims=True)
    image_latents = image_latents * functional.rsqrt(functional.reduce_sum(functional.square(image_latents), axis=-1, keepdims=True) + epsilon)

    similarity = functional.matmul(image_latents, text_latents, transpose_b=True)
    similarity = attention_layers.ExpLogitScale(axis=None, init_value=math.log(1 / 0.07), dtype="float32", name="temperature")(similarity)

    model = models.Model([image_input, text_input], similarity)
    model.run_prediction = RunPrediction(image_model, text_model, caption_tokenizer=caption_tokenizer)
    return model, image_model, text_model


def split_to_image_text_model(model):
    text_model = models.Model(model.inputs[1], model.get_layer("text_latents").output)

    image_latents_layer = model.get_layer("image_latents")
    image_output = image_latents_layer.input if isinstance(image_latents_layer, layers.Identity) else image_latents_layer.output
    image_model = models.Model(model.inputs[0], image_output)
    return image_model, text_model


def build_text_model_from_image_model(image_model, vocab_size=50257, text_dropout=0):
    from keras_cv_attention_models import model_surgery

    head_model, body_model, tail_model = model_surgery.split_model_to_head_body_tail_by_blocks(image_model)
    body_model = model_surgery.convert_to_dynamic_input_shape(body_model)

    image_input, image_head_output = head_model.inputs[0], head_model.outputs[0]
    image_body_output = body_model(image_head_output)
    image_output = tail_model(image_body_output)
    image_model = models.Model(image_input, image_output)  # Or models.Sequential([head_model, body_model, tail_model])

    body_input_rank = len(body_model.inputs[0].shape)
    assert body_input_rank == 3, "input for body has to be rank 3, got input_layer: {}, rank: {}".format(body_start_layer.name, body_input_rank)
    max_block_size, embedding_size = head_model.outputs[0].shape[1], body_model.inputs[0].shape[-1]

    text_input = layers.Input([None], dtype="int64")
    pos_idx = attention_layers.PositionalIndex(block_size=max_block_size, name="pos_idx")(text_input)
    tok_emb = layers.Embedding(vocab_size, embedding_size, name="wte")(text_input)
    pos_emb = layers.Embedding(max_block_size, embedding_size, name="wpe")(pos_idx)
    text_head_output = tok_emb + pos_emb
    text_head_output = layers.Dropout(dropout)(text_head_output) if text_dropout > 0 else text_head_output

    text_body_output = body_model(text_head_output)
    # text_output = text_model_index_header(text_input, text_body_output, latents_dim=image_model.outputs[0].shape[-1])
    text_model = models.Model(text_input, text_body_output, name=image_model.name + "_text")
    return image_model, text_model


class RunPrediction:
    def __init__(self, image_model, text_model, caption_tokenizer, softmax_scale=100, formatter="a {}", rescale_mode="torch"):
        self.image_model, self.text_model, self.caption_tokenizer = image_model, text_model, caption_tokenizer
        self.reset(softmax_scale=softmax_scale, formatter=formatter, rescale_mode=rescale_mode)

    def reset(self, softmax_scale=100, formatter="a {}", rescale_mode="torch"):
        self.formatter, self.text_labels, self.text_features = formatter, None, None
        self.softmax_scale = softmax_scale
        self.init_image_model_preprocess_input(rescale_mode)

    def init_image_model_preprocess_input(self, rescale_mode):
        if hasattr(self.image_model, "preprocess_input"):
            self.image_model.preprocess_input.set_rescale_mode(rescale_mode)
        else:
            self.image_model.preprocess_input = attention_layers.PreprocessInput(self.image_model.input_shape[1:], rescale_mode=rescale_mode)
        self.rescale_mode = rescale_mode

    def _if_init_text_features_(self, text_labels):
        if text_labels is None and self.text_labels is None:
            raise ValueError("[Error] text_labels has to be provided for the first time call. Format like ['cat', 'dog', ...]")

        if text_labels is not None:
            self.text_labels = text_labels if self.formatter is None else [self.formatter.format(ii) for ii in text_labels]

        if text_labels is not None or self.text_features is None:
            text_features = functional.stack([functional.convert_to_tensor(self.caption_tokenizer(ii), dtype="int64") for ii in self.text_labels])
            text_features = self.text_model(text_features)
            self.text_features = text_features / functional.norm(text_features, axis=-1, keepdims=True)

    def __call__(self, image, text_labels=None):
        self._if_init_text_features_(text_labels=text_labels)

        image_features = self.image_model(self.image_model.preprocess_input(image))
        image_features /= functional.norm(image_features, axis=-1, keepdims=True)
        return functional.softmax(self.softmax_scale * image_features @ functional.transpose(self.text_features))
